#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import argparse
import random

import subprocess
import sys
import time


def generate_runtimes(total_runtime):
    # combination of short runtimes and long runtimes, with heavier
    # weight on short runtimes
    possible_runtimes_sec = list(range(1, 10)) + list(range(1, 20)) + [100, 1000]
    runtimes = []
    while total_runtime > 0:
        chosen = random.choice(possible_runtimes_sec)
        chosen = min(chosen, total_runtime)
        runtimes.append(chosen)
        total_runtime -= chosen
    return runtimes


def main(args):
    runtimes = generate_runtimes(int(args.runtime_sec))
    print(
        "Going to execute write stress for " + str(runtimes)
    )  # noqa: E999 T25377293 Grandfathered in
    first_time = True

    for runtime in runtimes:
        kill = random.choice([False, True])

        cmd = "./write_stress --runtime_sec=" + ("-1" if kill else str(runtime))

        if len(args.db) > 0:
            cmd = cmd + " --db=" + args.db

        if first_time:
            first_time = False
        else:
            # use current db
            cmd = cmd + " --destroy_db=false"
        if random.choice([False, True]):
            cmd = cmd + " --delete_obsolete_files_with_fullscan=true"
        if random.choice([False, True]):
            cmd = cmd + " --low_open_files_mode=true"

        print(
            "Running write_stress for %d seconds (%s): %s"
            % (runtime, ("kill-mode" if kill else "clean-shutdown-mode"), cmd)
        )

        child = subprocess.Popen([cmd], shell=True)
        killtime = time.time() + runtime
        while not kill or time.time() < killtime:
            time.sleep(1)
            if child.poll() is not None:
                if child.returncode == 0:
                    break
                else:
                    print(
                        "ERROR: write_stress died with exitcode=%d\n" % child.returncode
                    )
                    sys.exit(1)
        if kill:
            child.kill()
        # breathe
        time.sleep(3)


if __name__ == "__main__":
    random.seed(time.time())
    parser = argparse.ArgumentParser(
        description="This script runs and kills \
        write_stress multiple times"
    )
    parser.add_argument("--runtime_sec", default="1000")
    parser.add_argument("--db", default="")
    args = parser.parse_args()
    main(args)



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import argparse

from advisor.db_config_optimizer import ConfigOptimizer
from advisor.db_log_parser import NO_COL_FAMILY
from advisor.db_options_parser import DatabaseOptions
from advisor.rule_parser import RulesSpec


CONFIG_OPT_NUM_ITER = 10


def main(args):
    # initialise the RulesSpec parser
    rule_spec_parser = RulesSpec(args.rules_spec)
    # initialise the benchmark runner
    bench_runner_module = __import__(
        args.benchrunner_module, fromlist=[args.benchrunner_class]
    )
    bench_runner_class = getattr(bench_runner_module, args.benchrunner_class)
    ods_args = {}
    if args.ods_client and args.ods_entity:
        ods_args["client_script"] = args.ods_client
        ods_args["entity"] = args.ods_entity
        if args.ods_key_prefix:
            ods_args["key_prefix"] = args.ods_key_prefix
    db_bench_runner = bench_runner_class(args.benchrunner_pos_args, ods_args)
    # initialise the database configuration
    db_options = DatabaseOptions(args.rocksdb_options, args.misc_options)
    # set the frequency at which stats are dumped in the LOG file and the
    # location of the LOG file.
    db_log_dump_settings = {
        "DBOptions.stats_dump_period_sec": {NO_COL_FAMILY: args.stats_dump_period_sec}
    }
    db_options.update_options(db_log_dump_settings)
    # initialise the configuration optimizer
    config_optimizer = ConfigOptimizer(
        db_bench_runner, db_options, rule_spec_parser, args.base_db_path
    )
    # run the optimiser to improve the database configuration for given
    # benchmarks, with the help of expert-specified rules
    final_db_options = config_optimizer.run()
    # generate the final rocksdb options file
    print(
        "Final configuration in: " + final_db_options.generate_options_config("final")
    )
    print("Final miscellaneous options: " + repr(final_db_options.get_misc_options()))


if __name__ == "__main__":
    """
    An example run of this tool from the command-line would look like:
    python3 -m advisor.config_optimizer_example
    --base_db_path=/tmp/rocksdbtest-155919/dbbench
    --rocksdb_options=temp/OPTIONS_boot.tmp --misc_options bloom_bits=2
    --rules_spec=advisor/rules.ini --stats_dump_period_sec=20
    --benchrunner_module=advisor.db_bench_runner
    --benchrunner_class=DBBenchRunner --benchrunner_pos_args ./../../db_bench
    readwhilewriting use_existing_db=true duration=90
    """
    parser = argparse.ArgumentParser(
        description="This script is used for\
        searching for a better database configuration"
    )
    parser.add_argument(
        "--rocksdb_options",
        required=True,
        type=str,
        help="path of the starting Rocksdb OPTIONS file",
    )
    # these are options that are column-family agnostic and are not yet
    # supported by the Rocksdb Options file: eg. bloom_bits=2
    parser.add_argument(
        "--misc_options",
        nargs="*",
        help="whitespace-separated list of options that are not supported "
        + "by the Rocksdb OPTIONS file, given in the "
        + '<option_name>=<option_value> format eg. "bloom_bits=2 '
        + 'rate_limiter_bytes_per_sec=128000000"',
    )
    parser.add_argument(
        "--base_db_path", required=True, type=str, help="path for the Rocksdb database"
    )
    parser.add_argument(
        "--rules_spec",
        required=True,
        type=str,
        help="path of the file containing the expert-specified Rules",
    )
    parser.add_argument(
        "--stats_dump_period_sec",
        required=True,
        type=int,
        help="the frequency (in seconds) at which STATISTICS are printed to "
        + "the Rocksdb LOG file",
    )
    # ODS arguments
    parser.add_argument("--ods_client", type=str, help="the ODS client binary")
    parser.add_argument(
        "--ods_entity",
        type=str,
        help="the servers for which the ODS stats need to be fetched",
    )
    parser.add_argument(
        "--ods_key_prefix",
        type=str,
        help="the prefix that needs to be attached to the keys of time "
        + "series to be fetched from ODS",
    )
    # benchrunner_module example: advisor.db_benchmark_client
    parser.add_argument(
        "--benchrunner_module",
        required=True,
        type=str,
        help="the module containing the BenchmarkRunner class to be used by "
        + "the Optimizer, example: advisor.db_bench_runner",
    )
    # benchrunner_class example: DBBenchRunner
    parser.add_argument(
        "--benchrunner_class",
        required=True,
        type=str,
        help="the name of the BenchmarkRunner class to be used by the "
        + "Optimizer, should be present in the module provided in the "
        + "benchrunner_module argument, example: DBBenchRunner",
    )
    parser.add_argument(
        "--benchrunner_pos_args",
        nargs="*",
        help="whitespace-separated positional arguments that are passed on "
        + "to the constructor of the BenchmarkRunner class provided in the "
        + 'benchrunner_class argument, example: "use_existing_db=true '
        + 'duration=900"',
    )
    args = parser.parse_args()
    main(args)



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import os
import unittest

from advisor.db_log_parser import DatabaseLogs, DataSource
from advisor.db_options_parser import DatabaseOptions
from advisor.rule_parser import RulesSpec

RuleToSuggestions = {
    "stall-too-many-memtables": ["inc-bg-flush", "inc-write-buffer"],
    "stall-too-many-L0": [
        "inc-max-subcompactions",
        "inc-max-bg-compactions",
        "inc-write-buffer-size",
        "dec-max-bytes-for-level-base",
        "inc-l0-slowdown-writes-trigger",
    ],
    "stop-too-many-L0": [
        "inc-max-bg-compactions",
        "inc-write-buffer-size",
        "inc-l0-stop-writes-trigger",
    ],
    "stall-too-many-compaction-bytes": [
        "inc-max-bg-compactions",
        "inc-write-buffer-size",
        "inc-hard-pending-compaction-bytes-limit",
        "inc-soft-pending-compaction-bytes-limit",
    ],
    "level0-level1-ratio": ["l0-l1-ratio-health-check"],
}


class TestAllRulesTriggered(unittest.TestCase):
    def setUp(self):
        # load the Rules
        this_path = os.path.abspath(os.path.dirname(__file__))
        ini_path = os.path.join(this_path, "input_files/triggered_rules.ini")
        self.db_rules = RulesSpec(ini_path)
        self.db_rules.load_rules_from_spec()
        self.db_rules.perform_section_checks()
        # load the data sources: LOG and OPTIONS
        log_path = os.path.join(this_path, "input_files/LOG-0")
        options_path = os.path.join(this_path, "input_files/OPTIONS-000005")
        db_options_parser = DatabaseOptions(options_path)
        self.column_families = db_options_parser.get_column_families()
        db_logs_parser = DatabaseLogs(log_path, self.column_families)
        self.data_sources = {
            DataSource.Type.DB_OPTIONS: [db_options_parser],
            DataSource.Type.LOG: [db_logs_parser],
        }

    def test_triggered_conditions(self):
        conditions_dict = self.db_rules.get_conditions_dict()
        rules_dict = self.db_rules.get_rules_dict()
        # Make sure none of the conditions is triggered beforehand
        for cond in conditions_dict.values():
            self.assertFalse(cond.is_triggered(), repr(cond))
        for rule in rules_dict.values():
            self.assertFalse(
                rule.is_triggered(conditions_dict, self.column_families), repr(rule)
            )

        # # Trigger the conditions as per the data sources.
        # trigger_conditions(, conditions_dict)

        # Get the set of rules that have been triggered
        triggered_rules = self.db_rules.get_triggered_rules(
            self.data_sources, self.column_families
        )

        # Make sure each condition and rule is triggered
        for cond in conditions_dict.values():
            if cond.get_data_source() is DataSource.Type.TIME_SERIES:
                continue
            self.assertTrue(cond.is_triggered(), repr(cond))

        for rule in rules_dict.values():
            self.assertIn(rule, triggered_rules)
            # Check the suggestions made by the triggered rules
            for sugg in rule.get_suggestions():
                self.assertIn(sugg, RuleToSuggestions[rule.name])

        for rule in triggered_rules:
            self.assertIn(rule, rules_dict.values())
            for sugg in RuleToSuggestions[rule.name]:
                self.assertIn(sugg, rule.get_suggestions())


class TestConditionsConjunctions(unittest.TestCase):
    def setUp(self):
        # load the Rules
        this_path = os.path.abspath(os.path.dirname(__file__))
        ini_path = os.path.join(this_path, "input_files/test_rules.ini")
        self.db_rules = RulesSpec(ini_path)
        self.db_rules.load_rules_from_spec()
        self.db_rules.perform_section_checks()
        # load the data sources: LOG and OPTIONS
        log_path = os.path.join(this_path, "input_files/LOG-1")
        options_path = os.path.join(this_path, "input_files/OPTIONS-000005")
        db_options_parser = DatabaseOptions(options_path)
        self.column_families = db_options_parser.get_column_families()
        db_logs_parser = DatabaseLogs(log_path, self.column_families)
        self.data_sources = {
            DataSource.Type.DB_OPTIONS: [db_options_parser],
            DataSource.Type.LOG: [db_logs_parser],
        }

    def test_condition_conjunctions(self):
        conditions_dict = self.db_rules.get_conditions_dict()
        rules_dict = self.db_rules.get_rules_dict()
        # Make sure none of the conditions is triggered beforehand
        for cond in conditions_dict.values():
            self.assertFalse(cond.is_triggered(), repr(cond))
        for rule in rules_dict.values():
            self.assertFalse(
                rule.is_triggered(conditions_dict, self.column_families), repr(rule)
            )

        # Trigger the conditions as per the data sources.
        self.db_rules.trigger_conditions(self.data_sources)

        # Check for the conditions
        conds_triggered = ["log-1-true", "log-2-true", "log-3-true"]
        conds_not_triggered = ["log-4-false"]
        for cond in conds_triggered:
            self.assertTrue(conditions_dict[cond].is_triggered(), repr(cond))
        for cond in conds_not_triggered:
            self.assertFalse(conditions_dict[cond].is_triggered(), repr(cond))

        # Check for the rules
        rules_triggered = ["multiple-conds-true"]
        rules_not_triggered = [
            "single-condition-false",
            "multiple-conds-one-false",
        ]
        for rule_name in rules_triggered:
            rule = rules_dict[rule_name]
            self.assertTrue(
                rule.is_triggered(conditions_dict, self.column_families), repr(rule)
            )
        for rule_name in rules_not_triggered:
            rule = rules_dict[rule_name]
            self.assertFalse(
                rule.is_triggered(conditions_dict, self.column_families), repr(rule)
            )


class TestSanityChecker(unittest.TestCase):
    def setUp(self):
        this_path = os.path.abspath(os.path.dirname(__file__))
        ini_path = os.path.join(this_path, "input_files/rules_err1.ini")
        db_rules = RulesSpec(ini_path)
        db_rules.load_rules_from_spec()
        self.rules_dict = db_rules.get_rules_dict()
        self.conditions_dict = db_rules.get_conditions_dict()
        self.suggestions_dict = db_rules.get_suggestions_dict()

    def test_rule_missing_suggestions(self):
        regex = ".*rule must have at least one suggestion.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.rules_dict["missing-suggestions"].perform_checks()

    def test_rule_missing_conditions(self):
        regex = ".*rule must have at least one condition.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.rules_dict["missing-conditions"].perform_checks()

    def test_condition_missing_regex(self):
        regex = ".*provide regex for log condition.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.conditions_dict["missing-regex"].perform_checks()

    def test_condition_missing_options(self):
        regex = ".*options missing in condition.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.conditions_dict["missing-options"].perform_checks()

    def test_condition_missing_expression(self):
        regex = ".*expression missing in condition.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.conditions_dict["missing-expression"].perform_checks()

    def test_suggestion_missing_option(self):
        regex = ".*provide option or description.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.suggestions_dict["missing-option"].perform_checks()

    def test_suggestion_missing_description(self):
        regex = ".*provide option or description.*"
        with self.assertRaisesRegex(ValueError, regex):
            self.suggestions_dict["missing-description"].perform_checks()


class TestParsingErrors(unittest.TestCase):
    def setUp(self):
        self.this_path = os.path.abspath(os.path.dirname(__file__))

    def test_condition_missing_source(self):
        ini_path = os.path.join(self.this_path, "input_files/rules_err2.ini")
        db_rules = RulesSpec(ini_path)
        regex = ".*provide source for condition.*"
        with self.assertRaisesRegex(NotImplementedError, regex):
            db_rules.load_rules_from_spec()

    def test_suggestion_missing_action(self):
        ini_path = os.path.join(self.this_path, "input_files/rules_err3.ini")
        db_rules = RulesSpec(ini_path)
        regex = ".*provide action for option.*"
        with self.assertRaisesRegex(ValueError, regex):
            db_rules.load_rules_from_spec()

    def test_section_no_name(self):
        ini_path = os.path.join(self.this_path, "input_files/rules_err4.ini")
        db_rules = RulesSpec(ini_path)
        regex = "Parsing error: needed section header:.*"
        with self.assertRaisesRegex(ValueError, regex):
            db_rules.load_rules_from_spec()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import os
import time
import unittest
from unittest.mock import MagicMock

from advisor.db_stats_fetcher import DatabasePerfContext, LogStatsParser
from advisor.db_timeseries_parser import NO_ENTITY
from advisor.rule_parser import Condition, TimeSeriesCondition


class TestLogStatsParser(unittest.TestCase):
    def setUp(self):
        this_path = os.path.abspath(os.path.dirname(__file__))
        stats_file = os.path.join(this_path, "input_files/log_stats_parser_keys_ts")
        # populate the keys_ts dictionary of LogStatsParser
        self.stats_dict = {NO_ENTITY: {}}
        with open(stats_file) as fp:
            for line in fp:
                stat_name = line.split(":")[0].strip()
                self.stats_dict[NO_ENTITY][stat_name] = {}
                token_list = line.split(":")[1].strip().split(",")
                for token in token_list:
                    timestamp = int(token.split()[0])
                    value = float(token.split()[1])
                    self.stats_dict[NO_ENTITY][stat_name][timestamp] = value
        self.log_stats_parser = LogStatsParser("dummy_log_file", 20)
        self.log_stats_parser.keys_ts = self.stats_dict

    def test_check_and_trigger_conditions_bursty(self):
        # mock fetch_timeseries() because 'keys_ts' has been pre-populated
        self.log_stats_parser.fetch_timeseries = MagicMock()
        # condition: bursty
        cond1 = Condition("cond-1")
        cond1 = TimeSeriesCondition.create(cond1)
        cond1.set_parameter("keys", "rocksdb.db.get.micros.p50")
        cond1.set_parameter("behavior", "bursty")
        cond1.set_parameter("window_sec", 40)
        cond1.set_parameter("rate_threshold", 0)
        self.log_stats_parser.check_and_trigger_conditions([cond1])
        expected_cond_trigger = {NO_ENTITY: {1530896440: 0.9767546362322214}}
        self.assertDictEqual(expected_cond_trigger, cond1.get_trigger())
        # ensure that fetch_timeseries() was called once
        self.log_stats_parser.fetch_timeseries.assert_called_once()

    def test_check_and_trigger_conditions_eval_agg(self):
        # mock fetch_timeseries() because 'keys_ts' has been pre-populated
        self.log_stats_parser.fetch_timeseries = MagicMock()
        # condition: evaluate_expression
        cond1 = Condition("cond-1")
        cond1 = TimeSeriesCondition.create(cond1)
        cond1.set_parameter("keys", "rocksdb.db.get.micros.p50")
        cond1.set_parameter("behavior", "evaluate_expression")
        keys = ["rocksdb.manifest.file.sync.micros.p99", "rocksdb.db.get.micros.p50"]
        cond1.set_parameter("keys", keys)
        cond1.set_parameter("aggregation_op", "latest")
        # condition evaluates to FALSE
        cond1.set_parameter("evaluate", "keys[0]-(keys[1]*100)>200")
        self.log_stats_parser.check_and_trigger_conditions([cond1])
        expected_cond_trigger = {NO_ENTITY: [1792.0, 15.9638]}
        self.assertIsNone(cond1.get_trigger())
        # condition evaluates to TRUE
        cond1.set_parameter("evaluate", "keys[0]-(keys[1]*100)<200")
        self.log_stats_parser.check_and_trigger_conditions([cond1])
        expected_cond_trigger = {NO_ENTITY: [1792.0, 15.9638]}
        self.assertDictEqual(expected_cond_trigger, cond1.get_trigger())
        # ensure that fetch_timeseries() was called
        self.log_stats_parser.fetch_timeseries.assert_called()

    def test_check_and_trigger_conditions_eval(self):
        # mock fetch_timeseries() because 'keys_ts' has been pre-populated
        self.log_stats_parser.fetch_timeseries = MagicMock()
        # condition: evaluate_expression
        cond1 = Condition("cond-1")
        cond1 = TimeSeriesCondition.create(cond1)
        cond1.set_parameter("keys", "rocksdb.db.get.micros.p50")
        cond1.set_parameter("behavior", "evaluate_expression")
        keys = ["rocksdb.manifest.file.sync.micros.p99", "rocksdb.db.get.micros.p50"]
        cond1.set_parameter("keys", keys)
        cond1.set_parameter("evaluate", "keys[0]-(keys[1]*100)>500")
        self.log_stats_parser.check_and_trigger_conditions([cond1])
        expected_trigger = {
            NO_ENTITY: {
                1530896414: [9938.0, 16.31508],
                1530896440: [9938.0, 16.346602],
                1530896466: [9938.0, 16.284669],
                1530896492: [9938.0, 16.16005],
            }
        }
        self.assertDictEqual(expected_trigger, cond1.get_trigger())
        self.log_stats_parser.fetch_timeseries.assert_called_once()


class TestDatabasePerfContext(unittest.TestCase):
    def test_unaccumulate_metrics(self):
        perf_dict = {
            "user_key_comparison_count": 675903942,
            "block_cache_hit_count": 830086,
        }
        timestamp = int(time.time())
        perf_ts = {}
        for key in perf_dict:
            perf_ts[key] = {}
            start_val = perf_dict[key]
            for ix in range(5):
                perf_ts[key][timestamp + (ix * 10)] = start_val + (2 * ix * ix)
        db_perf_context = DatabasePerfContext(perf_ts, 10, True)
        timestamps = [timestamp + (ix * 10) for ix in range(1, 5, 1)]
        values = [val for val in range(2, 15, 4)]
        inner_dict = {timestamps[ix]: values[ix] for ix in range(4)}
        expected_keys_ts = {
            NO_ENTITY: {
                "user_key_comparison_count": inner_dict,
                "block_cache_hit_count": inner_dict,
            }
        }
        self.assertDictEqual(expected_keys_ts, db_perf_context.keys_ts)



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import argparse

from advisor.db_log_parser import DatabaseLogs, DataSource
from advisor.db_options_parser import DatabaseOptions
from advisor.db_stats_fetcher import LogStatsParser, OdsStatsFetcher
from advisor.rule_parser import RulesSpec


def main(args):
    # initialise the RulesSpec parser
    rule_spec_parser = RulesSpec(args.rules_spec)
    rule_spec_parser.load_rules_from_spec()
    rule_spec_parser.perform_section_checks()
    # initialize the DatabaseOptions object
    db_options = DatabaseOptions(args.rocksdb_options)
    # Create DatabaseLogs object
    db_logs = DatabaseLogs(args.log_files_path_prefix, db_options.get_column_families())
    # Create the Log STATS object
    db_log_stats = LogStatsParser(
        args.log_files_path_prefix, args.stats_dump_period_sec
    )
    data_sources = {
        DataSource.Type.DB_OPTIONS: [db_options],
        DataSource.Type.LOG: [db_logs],
        DataSource.Type.TIME_SERIES: [db_log_stats],
    }
    if args.ods_client:
        data_sources[DataSource.Type.TIME_SERIES].append(
            OdsStatsFetcher(
                args.ods_client,
                args.ods_entity,
                args.ods_tstart,
                args.ods_tend,
                args.ods_key_prefix,
            )
        )
    triggered_rules = rule_spec_parser.get_triggered_rules(
        data_sources, db_options.get_column_families()
    )
    rule_spec_parser.print_rules(triggered_rules)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Use this script to get\
        suggestions for improving Rocksdb performance."
    )
    parser.add_argument(
        "--rules_spec",
        required=True,
        type=str,
        help="path of the file containing the expert-specified Rules",
    )
    parser.add_argument(
        "--rocksdb_options",
        required=True,
        type=str,
        help="path of the starting Rocksdb OPTIONS file",
    )
    parser.add_argument(
        "--log_files_path_prefix",
        required=True,
        type=str,
        help="path prefix of the Rocksdb LOG files",
    )
    parser.add_argument(
        "--stats_dump_period_sec",
        required=True,
        type=int,
        help="the frequency (in seconds) at which STATISTICS are printed to "
        + "the Rocksdb LOG file",
    )
    # ODS arguments
    parser.add_argument("--ods_client", type=str, help="the ODS client binary")
    parser.add_argument(
        "--ods_entity",
        type=str,
        help="the servers for which the ODS stats need to be fetched",
    )
    parser.add_argument(
        "--ods_key_prefix",
        type=str,
        help="the prefix that needs to be attached to the keys of time "
        + "series to be fetched from ODS",
    )
    parser.add_argument(
        "--ods_tstart", type=int, help="start time of timeseries to be fetched from ODS"
    )
    parser.add_argument(
        "--ods_tend", type=int, help="end time of timeseries to be fetched from ODS"
    )
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import glob

import os
import os.path
import re
import shutil
import subprocess
import tempfile
import time
import unittest


def my_check_output(*popenargs, **kwargs):
    """
    If we had python 2.7, we should simply use subprocess.check_output.
    This is a stop-gap solution for python 2.6
    """
    if "stdout" in kwargs:
        raise ValueError("stdout argument not allowed, it will be overridden.")
    process = subprocess.Popen(
        stderr=subprocess.PIPE, stdout=subprocess.PIPE, *popenargs, **kwargs
    )
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise Exception("Exit code is not 0.  It is %d.  Command: %s" % (retcode, cmd))
    return output.decode("utf-8")


def run_err_null(cmd):
    return os.system(cmd + " 2>/dev/null ")


class LDBTestCase(unittest.TestCase):
    def setUp(self):
        self.TMP_DIR = tempfile.mkdtemp(prefix="ldb_test_")
        self.DB_NAME = "testdb"

    def tearDown(self):
        assert (
            self.TMP_DIR.strip() != "/"
            and self.TMP_DIR.strip() != "/tmp"
            and self.TMP_DIR.strip() != "/tmp/"
        )  # Just some paranoia

        shutil.rmtree(self.TMP_DIR)

    def dbParam(self, dbName):
        return "--db=%s" % os.path.join(self.TMP_DIR, dbName)

    def assertRunOKFull(
        self, params, expectedOutput, unexpected=False, isPattern=False
    ):
        """
        All command-line params must be specified.
        Allows full flexibility in testing; for example: missing db param.
        """
        output = my_check_output(
            './ldb %s |grep -v "Created bg thread"' % params, shell=True
        )
        if not unexpected:
            if isPattern:
                self.assertNotEqual(expectedOutput.search(output.strip()), None)
            else:
                self.assertEqual(output.strip(), expectedOutput.strip())
        else:
            if isPattern:
                self.assertEqual(expectedOutput.search(output.strip()), None)
            else:
                self.assertNotEqual(output.strip(), expectedOutput.strip())

    def assertRunFAILFull(self, params):
        """
        All command-line params must be specified.
        Allows full flexibility in testing; for example: missing db param.
        """
        try:

            my_check_output(
                './ldb %s >/dev/null 2>&1 |grep -v "Created bg \
                thread"'
                % params,
                shell=True,
            )
        except Exception:
            return
        self.fail(
            "Exception should have been raised for command with params: %s" % params
        )

    def assertRunOK(self, params, expectedOutput, unexpected=False):
        """
        Uses the default test db.
        """
        self.assertRunOKFull(
            "{} {}".format(self.dbParam(self.DB_NAME), params), expectedOutput, unexpected
        )

    def assertRunFAIL(self, params):
        """
        Uses the default test db.
        """
        self.assertRunFAILFull("{} {}".format(self.dbParam(self.DB_NAME), params))

    def testSimpleStringPutGet(self):
        print("Running testSimpleStringPutGet...")
        self.assertRunFAIL("put x1 y1")
        self.assertRunOK("put --create_if_missing x1 y1", "OK")
        self.assertRunOK("get x1", "y1")
        self.assertRunFAIL("get x2")

        self.assertRunOK("put x2 y2", "OK")
        self.assertRunOK("get x1", "y1")
        self.assertRunOK("get x2", "y2")
        self.assertRunOK("multi_get x1 x2", "x1 ==> y1\nx2 ==> y2")
        self.assertRunOK("get x3", "Key not found")
        self.assertRunOK("multi_get x3", "Key not found: x3")

        self.assertRunFAIL("put_entity x4")
        self.assertRunFAIL("put_entity x4 cv1")
        self.assertRunOK("put_entity x4 :cv1", "OK")
        self.assertRunOK("get_entity x4", ":cv1")

        self.assertRunOK("put_entity x5 cn1:cv1 cn2:cv2", "OK")
        self.assertRunOK("get_entity x5", "cn1:cv1 cn2:cv2")

        self.assertRunOK(
            "scan --from=x1 --to=z",
            "x1 ==> y1\nx2 ==> y2\nx4 ==> cv1\nx5 ==> cn1:cv1 cn2:cv2",
        )
        self.assertRunOK("put x3 y3", "OK")

        self.assertRunOK(
            "scan --from=x1 --to=z",
            "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> cv1\nx5 ==> cn1:cv1 cn2:cv2",
        )
        self.assertRunOK(
            "scan",
            "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> cv1\nx5 ==> cn1:cv1 cn2:cv2",
        )
        self.assertRunOK(
            "scan --from=x",
            "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> cv1\nx5 ==> cn1:cv1 cn2:cv2",
        )

        self.assertRunOK("scan --to=x2", "x1 ==> y1")
        self.assertRunOK("scan --from=x1 --to=z --max_keys=1", "x1 ==> y1")
        self.assertRunOK("scan --from=x1 --to=z --max_keys=2", "x1 ==> y1\nx2 ==> y2")

        self.assertRunOK("delete x4", "OK")
        self.assertRunOK("delete x5", "OK")

        self.assertRunOK(
            "scan --from=x1 --to=z --max_keys=3", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3"
        )
        self.assertRunOK(
            "scan --from=x1 --to=z --max_keys=4", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3"
        )
        self.assertRunOK("scan --from=x1 --to=x2", "x1 ==> y1")
        self.assertRunOK("scan --from=x2 --to=x4", "x2 ==> y2\nx3 ==> y3")
        self.assertRunFAIL("scan --from=x4 --to=z")  # No results => FAIL
        self.assertRunFAIL("scan --from=x1 --to=z --max_keys=foo")

        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3")

        self.assertRunOK("delete x1", "OK")
        self.assertRunOK("scan", "x2 ==> y2\nx3 ==> y3")

        self.assertRunOK("delete NonExistentKey", "OK")
        # It is weird that GET and SCAN raise exception for
        # non-existent key, while delete does not

        self.assertRunOK("checkconsistency", "OK")

    def dumpDb(self, params, dumpFile):
        return 0 == run_err_null("./ldb dump {} > {}".format(params, dumpFile))

    def loadDb(self, params, dumpFile):
        return 0 == run_err_null("cat {} | ./ldb load {}".format(dumpFile, params))

    def writeExternSst(self, params, inputDumpFile, outputSst):
        return 0 == run_err_null(
            "cat {} | ./ldb write_extern_sst {} {}".format(inputDumpFile, outputSst, params)
        )

    def ingestExternSst(self, params, inputSst):
        return 0 == run_err_null("./ldb ingest_extern_sst {} {}".format(inputSst, params))

    def testStringBatchPut(self):
        print("Running testStringBatchPut...")
        self.assertRunOK("batchput x1 y1 --create_if_missing", "OK")
        self.assertRunOK("scan", "x1 ==> y1")
        self.assertRunOK('batchput x2 y2 x3 y3 "x4 abc" "y4 xyz"', "OK")
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 abc ==> y4 xyz")
        self.assertRunFAIL("batchput")
        self.assertRunFAIL("batchput k1")
        self.assertRunFAIL("batchput k1 v1 k2")

    def testBlobBatchPut(self):
        print("Running testBlobBatchPut...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("batchput x1 y1 --create_if_missing --enable_blob_files", "OK")
        self.assertRunOK("scan", "x1 ==> y1")
        self.assertRunOK(
            'batchput --enable_blob_files x2 y2 x3 y3 "x4 abc" "y4 xyz"', "OK"
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 abc ==> y4 xyz")

        blob_files = self.getBlobFiles(dbPath)
        self.assertTrue(len(blob_files) >= 1)

    def testBlobPut(self):
        print("Running testBlobPut...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put --create_if_missing --enable_blob_files x1 y1", "OK")
        self.assertRunOK("get x1", "y1")
        self.assertRunOK("put --enable_blob_files x2 y2", "OK")
        self.assertRunOK("get x1", "y1")
        self.assertRunOK("get x2", "y2")
        self.assertRunFAIL("get x3")

        blob_files = self.getBlobFiles(dbPath)
        self.assertTrue(len(blob_files) >= 1)

    def testBlobStartingLevel(self):
        print("Running testBlobStartingLevel...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK(
            "put --create_if_missing --enable_blob_files --blob_file_starting_level=10 x1 y1",
            "OK",
        )
        self.assertRunOK("get x1", "y1")

        blob_files = self.getBlobFiles(dbPath)
        self.assertTrue(len(blob_files) == 0)

        self.assertRunOK(
            "put --enable_blob_files --blob_file_starting_level=0 x2 y2", "OK"
        )
        self.assertRunOK("get x1", "y1")
        self.assertRunOK("get x2", "y2")
        self.assertRunFAIL("get x3")

        blob_files = self.getBlobFiles(dbPath)
        self.assertTrue(len(blob_files) >= 1)

    def testCountDelimDump(self):
        print("Running testCountDelimDump...")
        self.assertRunOK("batchput x.1 x1 --create_if_missing", "OK")
        self.assertRunOK("batchput y.abc abc y.2 2 z.13c pqr", "OK")
        self.assertRunOK(
            "dump --count_delim",
            "x => count:1\tsize:5\ny => count:2\tsize:12\nz => count:1\tsize:8",
        )
        self.assertRunOK(
            'dump --count_delim="."',
            "x => count:1\tsize:5\ny => count:2\tsize:12\nz => count:1\tsize:8",
        )
        self.assertRunOK("batchput x,2 x2 x,abc xabc", "OK")
        self.assertRunOK(
            'dump --count_delim=","',
            "x => count:2\tsize:14\nx.1 => count:1\tsize:5\ny.2 => count:1\tsize:4\ny.abc => count:1\tsize:8\nz.13c => count:1\tsize:8",
        )

    def testCountDelimIDump(self):
        print("Running testCountDelimIDump...")
        self.assertRunOK("batchput x.1 x1 --create_if_missing", "OK")
        self.assertRunOK("batchput y.abc abc y.2 2 z.13c pqr", "OK")
        self.assertRunOK(
            "idump --count_delim",
            "x => count:1\tsize:5\ny => count:2\tsize:12\nz => count:1\tsize:8",
        )
        self.assertRunOK(
            'idump --count_delim="."',
            "x => count:1\tsize:5\ny => count:2\tsize:12\nz => count:1\tsize:8",
        )
        self.assertRunOK("batchput x,2 x2 x,abc xabc", "OK")
        self.assertRunOK(
            'idump --count_delim=","',
            "x => count:2\tsize:14\nx.1 => count:1\tsize:5\ny.2 => count:1\tsize:4\ny.abc => count:1\tsize:8\nz.13c => count:1\tsize:8",
        )

    def testInvalidCmdLines(self):
        print("Running testInvalidCmdLines...")
        # db not specified
        self.assertRunFAILFull("put 0x6133 0x6233 --hex --create_if_missing")
        # No param called he
        self.assertRunFAIL("put 0x6133 0x6233 --he --create_if_missing")
        # max_keys is not applicable for put
        self.assertRunFAIL("put 0x6133 0x6233 --max_keys=1 --create_if_missing")
        # hex has invalid boolean value

    def testHexPutGet(self):
        print("Running testHexPutGet...")
        self.assertRunOK("put a1 b1 --create_if_missing", "OK")
        self.assertRunOK("scan", "a1 ==> b1")
        self.assertRunOK("scan --hex", "0x6131 ==> 0x6231")
        self.assertRunFAIL("put --hex 6132 6232")
        self.assertRunOK("put --hex 0x6132 0x6232", "OK")
        self.assertRunOK("scan --hex", "0x6131 ==> 0x6231\n0x6132 ==> 0x6232")
        self.assertRunOK("scan", "a1 ==> b1\na2 ==> b2")
        self.assertRunOK("get a1", "b1")
        self.assertRunOK("get --hex 0x6131", "0x6231")
        self.assertRunOK("get a2", "b2")
        self.assertRunOK("get --hex 0x6132", "0x6232")
        self.assertRunOK("multi_get --hex 0x6131 0x6132", "0x6131 ==> 0x6231\n0x6132 ==> 0x6232")
        self.assertRunOK("multi_get --hex 0x6131 0xBEEF", "0x6131 ==> 0x6231\nKey not found: 0xBEEF")
        self.assertRunOK("get --key_hex 0x6132", "b2")
        self.assertRunOK("get --key_hex --value_hex 0x6132", "0x6232")
        self.assertRunOK("get --value_hex a2", "0x6232")
        self.assertRunOK(
            "scan --key_hex --value_hex", "0x6131 ==> 0x6231\n0x6132 ==> 0x6232"
        )
        self.assertRunOK(
            "scan --hex --from=0x6131 --to=0x6133",
            "0x6131 ==> 0x6231\n0x6132 ==> 0x6232",
        )
        self.assertRunOK("scan --hex --from=0x6131 --to=0x6132", "0x6131 ==> 0x6231")
        self.assertRunOK("scan --key_hex", "0x6131 ==> b1\n0x6132 ==> b2")
        self.assertRunOK("scan --value_hex", "a1 ==> 0x6231\na2 ==> 0x6232")
        self.assertRunOK("batchput --hex 0x6133 0x6233 0x6134 0x6234", "OK")
        self.assertRunOK("scan", "a1 ==> b1\na2 ==> b2\na3 ==> b3\na4 ==> b4")
        self.assertRunOK("delete --hex 0x6133", "OK")
        self.assertRunOK("scan", "a1 ==> b1\na2 ==> b2\na4 ==> b4")
        self.assertRunOK("checkconsistency", "OK")

    def testTtlPutGet(self):
        print("Running testTtlPutGet...")
        self.assertRunOK("put a1 b1 --ttl --create_if_missing", "OK")
        self.assertRunOK("scan --hex", "0x6131 ==> 0x6231", True)
        self.assertRunOK("dump --ttl ", "a1 ==> b1", True)
        self.assertRunOK("dump --hex --ttl ", "0x6131 ==> 0x6231\nKeys in range: 1")
        self.assertRunOK("scan --hex --ttl", "0x6131 ==> 0x6231")
        self.assertRunOK("get --value_hex a1", "0x6231", True)
        self.assertRunOK("get --ttl a1", "b1")
        self.assertRunOK("put a3 b3 --create_if_missing", "OK")
        # fails because timstamp's length is greater than value's
        self.assertRunFAIL("get --ttl a3")
        self.assertRunOK("checkconsistency", "OK")

    def testInvalidCmdLines(self):  # noqa: F811 T25377293 Grandfathered in
        print("Running testInvalidCmdLines...")
        # db not specified
        self.assertRunFAILFull("put 0x6133 0x6233 --hex --create_if_missing")
        # No param called he
        self.assertRunFAIL("put 0x6133 0x6233 --he --create_if_missing")
        # max_keys is not applicable for put
        self.assertRunFAIL("put 0x6133 0x6233 --max_keys=1 --create_if_missing")
        # hex has invalid boolean value
        self.assertRunFAIL("put 0x6133 0x6233 --hex=Boo --create_if_missing")

    def testDumpLoad(self):
        print("Running testDumpLoad...")
        self.assertRunOK("batchput --create_if_missing x1 y1 x2 y2 x3 y3 x4 y4", "OK")
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")
        origDbPath = os.path.join(self.TMP_DIR, self.DB_NAME)

        # Dump and load without any additional params specified
        dumpFilePath = os.path.join(self.TMP_DIR, "dump1")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump1")
        self.assertTrue(self.dumpDb("--db=%s" % origDbPath, dumpFilePath))
        self.assertTrue(
            self.loadDb("--db=%s --create_if_missing" % loadedDbPath, dumpFilePath)
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )

        # Dump and load in hex
        dumpFilePath = os.path.join(self.TMP_DIR, "dump2")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump2")
        self.assertTrue(self.dumpDb("--db=%s --hex" % origDbPath, dumpFilePath))
        self.assertTrue(
            self.loadDb(
                "--db=%s --hex --create_if_missing" % loadedDbPath, dumpFilePath
            )
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )

        # Dump only a portion of the key range
        dumpFilePath = os.path.join(self.TMP_DIR, "dump3")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump3")
        self.assertTrue(
            self.dumpDb("--db=%s --from=x1 --to=x3" % origDbPath, dumpFilePath)
        )
        self.assertTrue(
            self.loadDb("--db=%s --create_if_missing" % loadedDbPath, dumpFilePath)
        )
        self.assertRunOKFull("scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2")

        # Dump upto max_keys rows
        dumpFilePath = os.path.join(self.TMP_DIR, "dump4")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump4")
        self.assertTrue(self.dumpDb("--db=%s --max_keys=3" % origDbPath, dumpFilePath))
        self.assertTrue(
            self.loadDb("--db=%s --create_if_missing" % loadedDbPath, dumpFilePath)
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3"
        )

        # Load into an existing db, create_if_missing is not specified
        self.assertTrue(self.dumpDb("--db=%s" % origDbPath, dumpFilePath))
        self.assertTrue(self.loadDb("--db=%s" % loadedDbPath, dumpFilePath))
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )

        # Dump and load with WAL disabled
        dumpFilePath = os.path.join(self.TMP_DIR, "dump5")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump5")
        self.assertTrue(self.dumpDb("--db=%s" % origDbPath, dumpFilePath))
        self.assertTrue(
            self.loadDb(
                "--db=%s --disable_wal --create_if_missing" % loadedDbPath, dumpFilePath
            )
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )

        # Dump and load with lots of extra params specified
        extraParams = " ".join(
            [
                "--bloom_bits=14",
                "--block_size=1024",
                "--auto_compaction=true",
                "--write_buffer_size=4194304",
                "--file_size=2097152",
            ]
        )
        dumpFilePath = os.path.join(self.TMP_DIR, "dump6")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump6")
        self.assertTrue(
            self.dumpDb("--db={} {}".format(origDbPath, extraParams), dumpFilePath)
        )
        self.assertTrue(
            self.loadDb(
                "--db={} {} --create_if_missing".format(loadedDbPath, extraParams),
                dumpFilePath,
            )
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )

        # Dump with count_only
        dumpFilePath = os.path.join(self.TMP_DIR, "dump7")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump7")
        self.assertTrue(self.dumpDb("--db=%s --count_only" % origDbPath, dumpFilePath))
        self.assertTrue(
            self.loadDb("--db=%s --create_if_missing" % loadedDbPath, dumpFilePath)
        )
        # DB should have atleast one value for scan to work
        self.assertRunOKFull("put --db=%s k1 v1" % loadedDbPath, "OK")
        self.assertRunOKFull("scan --db=%s" % loadedDbPath, "k1 ==> v1")

        # Dump command fails because of typo in params
        dumpFilePath = os.path.join(self.TMP_DIR, "dump8")
        self.assertFalse(
            self.dumpDb("--db=%s --create_if_missing" % origDbPath, dumpFilePath)
        )

        # Dump and load with BlobDB enabled
        blobParams = " ".join(
            ["--enable_blob_files", "--min_blob_size=1", "--blob_file_size=2097152"]
        )
        dumpFilePath = os.path.join(self.TMP_DIR, "dump9")
        loadedDbPath = os.path.join(self.TMP_DIR, "loaded_from_dump9")
        self.assertTrue(self.dumpDb("--db=%s" % (origDbPath), dumpFilePath))
        self.assertTrue(
            self.loadDb(
                "--db=%s %s --create_if_missing --disable_wal"
                % (loadedDbPath, blobParams),
                dumpFilePath,
            )
        )
        self.assertRunOKFull(
            "scan --db=%s" % loadedDbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )
        blob_files = self.getBlobFiles(loadedDbPath)
        self.assertTrue(len(blob_files) >= 1)

    def testIDumpBasics(self):
        print("Running testIDumpBasics...")
        self.assertRunOK("put a val --create_if_missing", "OK")
        self.assertRunOK("put b val", "OK")
        self.assertRunOK(
            "idump",
            "'a' seq:1, type:1 => val\n"
            "'b' seq:2, type:1 => val\nInternal keys in range: 2",
        )
        self.assertRunOK(
            "idump --input_key_hex --from={} --to={}".format(hex(ord("a")), hex(ord("b"))),
            "'a' seq:1, type:1 => val\nInternal keys in range: 1",
        )

    def testIDumpDecodeBlobIndex(self):
        print("Running testIDumpDecodeBlobIndex...")
        self.assertRunOK("put a val --create_if_missing", "OK")
        self.assertRunOK("put b val --enable_blob_files", "OK")

        # Pattern to expect from dump with decode_blob_index flag enabled.
        regex = r".*\[blob ref\].*"
        expected_pattern = re.compile(regex)
        cmd = "idump %s --decode_blob_index"
        self.assertRunOKFull(
            (cmd) % (self.dbParam(self.DB_NAME)),
            expected_pattern,
            unexpected=False,
            isPattern=True,
        )

    def testMiscAdminTask(self):
        print("Running testMiscAdminTask...")
        # These tests need to be improved; for example with asserts about
        # whether compaction or level reduction actually took place.
        self.assertRunOK("batchput --create_if_missing x1 y1 x2 y2 x3 y3 x4 y4", "OK")
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")
        origDbPath = os.path.join(self.TMP_DIR, self.DB_NAME)

        self.assertTrue(0 == run_err_null("./ldb compact --db=%s" % origDbPath))
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

        self.assertTrue(
            0 == run_err_null("./ldb reduce_levels --db=%s --new_levels=2" % origDbPath)
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

        self.assertTrue(
            0 == run_err_null("./ldb reduce_levels --db=%s --new_levels=3" % origDbPath)
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

        self.assertTrue(
            0 == run_err_null("./ldb compact --db=%s --from=x1 --to=x3" % origDbPath)
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

        self.assertTrue(
            0
            == run_err_null(
                "./ldb compact --db=%s --hex --from=0x6131 --to=0x6134" % origDbPath
            )
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

        # TODO(dilip): Not sure what should be passed to WAL.Currently corrupted.
        self.assertTrue(
            0
            == run_err_null(
                "./ldb dump_wal --db=%s --walfile=%s --header"
                % (origDbPath, origDbPath)
            )
        )
        self.assertRunOK("scan", "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4")

    def testCheckConsistency(self):
        print("Running testCheckConsistency...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put x1 y1 --create_if_missing", "OK")
        self.assertRunOK("put x2 y2", "OK")
        self.assertRunOK("get x1", "y1")
        self.assertRunOK("checkconsistency", "OK")

        sstFilePath = my_check_output(
            "ls %s" % os.path.join(dbPath, "*.sst"), shell=True
        )

        # Modify the file
        my_check_output("echo 'evil' > %s" % sstFilePath, shell=True)
        self.assertRunFAIL("checkconsistency")

        # Delete the file
        my_check_output("rm -f %s" % sstFilePath, shell=True)
        self.assertRunFAIL("checkconsistency")

    def dumpLiveFiles(self, params, dumpFile):
        return 0 == run_err_null("./ldb dump_live_files {} > {}".format(params, dumpFile))

    def testDumpLiveFiles(self):
        print("Running testDumpLiveFiles...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put x1 y1 --create_if_missing", "OK")
        self.assertRunOK("put x2 y2 --enable_blob_files", "OK")
        dumpFilePath = os.path.join(self.TMP_DIR, "dump1")
        self.assertTrue(self.dumpLiveFiles("--db=%s" % dbPath, dumpFilePath))
        self.assertRunOK("delete x1", "OK")
        self.assertRunOK("put x3 y3", "OK")
        dumpFilePath = os.path.join(self.TMP_DIR, "dump2")

        # Test that if the user provides a db path that ends with
        # a slash '/', there is no double (or more!) slashes in the
        # SST and manifest file names.

        # Add a '/' at the end of dbPath (which normally shouldnt contain any)
        if dbPath[-1] != "/":
            dbPath += "/"

        # Call the dump_live_files function with the edited dbPath name.
        self.assertTrue(
            self.dumpLiveFiles(
                "--db=%s --decode_blob_index --dump_uncompressed_blobs" % dbPath,
                dumpFilePath,
            )
        )

        # Investigate the output
        with open(dumpFilePath) as tmp:
            data = tmp.read()

        # Check that all the SST filenames have a correct full path (no multiple '/').
        sstFileList = re.findall(r"%s.*\d+.sst" % dbPath, data)
        self.assertTrue(len(sstFileList) >= 1)
        for sstFilename in sstFileList:
            filenumber = re.findall(r"\d+.sst", sstFilename)[0]
            self.assertEqual(sstFilename, dbPath + filenumber)

        # Check that all the Blob filenames have a correct full path (no multiple '/').
        blobFileList = re.findall(r"%s.*\d+.blob" % dbPath, data)
        self.assertTrue(len(blobFileList) >= 1)
        for blobFilename in blobFileList:
            filenumber = re.findall(r"\d+.blob", blobFilename)[0]
            self.assertEqual(blobFilename, dbPath + filenumber)

        # Check that all the manifest filenames
        # have a correct full path (no multiple '/').
        manifestFileList = re.findall(r"%s.*MANIFEST-\d+" % dbPath, data)
        self.assertTrue(len(manifestFileList) >= 1)
        for manifestFilename in manifestFileList:
            filenumber = re.findall(r"(?<=MANIFEST-)\d+", manifestFilename)[0]
            self.assertEqual(manifestFilename, dbPath + "MANIFEST-" + filenumber)

        # Check that the blob file index is decoded.
        decodedBlobIndex = re.findall(r"\[blob ref\]", data)
        self.assertTrue(len(decodedBlobIndex) >= 1)

    def listLiveFilesMetadata(self, params, dumpFile):
        return 0 == run_err_null(
            "./ldb list_live_files_metadata {} > {}".format(params, dumpFile)
        )

    def testListLiveFilesMetadata(self):
        print("Running testListLiveFilesMetadata...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put x1 y1 --create_if_missing", "OK")
        self.assertRunOK("put x2 y2", "OK")

        # Compare the SST filename and the level of list_live_files_metadata
        # with the data collected from dump_live_files.
        dumpFilePath1 = os.path.join(self.TMP_DIR, "dump1")
        self.assertTrue(self.dumpLiveFiles("--db=%s" % dbPath, dumpFilePath1))
        dumpFilePath2 = os.path.join(self.TMP_DIR, "dump2")
        self.assertTrue(
            self.listLiveFilesMetadata(
                "--sort_by_filename --db=%s" % dbPath, dumpFilePath2
            )
        )

        # Collect SST filename and level from dump_live_files
        with open(dumpFilePath1) as tmp:
            data = tmp.read()
            filename1 = re.findall(r".*\d+\.sst", data)[0]
            level1 = re.findall(r"level:\d+", data)[0].split(":")[1]

        # Collect SST filename and level from list_live_files_metadata
        with open(dumpFilePath2) as tmp:
            data = tmp.read()
            filename2 = re.findall(r".*\d+\.sst", data)[0]
            level2 = re.findall(r"level \d+", data)[0].split(" ")[1]

        # Assert equality between filenames and levels.
        self.assertEqual(filename1, filename2)
        self.assertEqual(level1, level2)

        # Create multiple column families and compare the output
        # of list_live_files_metadata with dump_live_files once again.
        # Create new CF, and insert data:
        self.assertRunOK("create_column_family mycol1", "OK")
        self.assertRunOK("put --column_family=mycol1 v1 v2", "OK")
        self.assertRunOK("create_column_family mycol2", "OK")
        self.assertRunOK("put --column_family=mycol2 h1 h2", "OK")
        self.assertRunOK("put --column_family=mycol2 h3 h4", "OK")

        # Call dump_live_files and list_live_files_metadata
        # and pipe the output to compare them later.
        dumpFilePath3 = os.path.join(self.TMP_DIR, "dump3")
        self.assertTrue(self.dumpLiveFiles("--db=%s" % dbPath, dumpFilePath3))
        dumpFilePath4 = os.path.join(self.TMP_DIR, "dump4")
        self.assertTrue(
            self.listLiveFilesMetadata(
                "--sort_by_filename --db=%s" % dbPath, dumpFilePath4
            )
        )

        # dump_live_files:
        # parse the output and create a map:
        # [key: sstFilename]->[value:[LSM level, Column Family Name]]
        referenceMap = {}
        with open(dumpFilePath3) as tmp:
            data = tmp.read()
            # Note: the following regex are contingent on what the
            # dump_live_files outputs.
            namesAndLevels = re.findall(r"\d+.sst level:\d+", data)
            cfs = re.findall(r"(?<=column family name=)\w+", data)
            # re.findall should not reorder the data.
            # Therefore namesAndLevels[i] matches the data from cfs[i].
            for count, nameAndLevel in enumerate(namesAndLevels):
                sstFilename = re.findall(r"\d+.sst", nameAndLevel)[0]
                sstLevel = re.findall(r"(?<=level:)\d+", nameAndLevel)[0]
                cf = cfs[count]
                referenceMap[sstFilename] = [sstLevel, cf]

        # list_live_files_metadata:
        # parse the output and create a map:
        # [key: sstFilename]->[value:[LSM level, Column Family Name]]
        testMap = {}
        with open(dumpFilePath4) as tmp:
            data = tmp.read()
            # Since for each SST file, all the information is contained
            # on one line, the parsing is easy to perform and relies on
            # the appearance of an "00xxx.sst" pattern.
            sstLines = re.findall(r".*\d+.sst.*", data)
            for line in sstLines:
                sstFilename = re.findall(r"\d+.sst", line)[0]
                sstLevel = re.findall(r"(?<=level )\d+", line)[0]
                cf = re.findall(r"(?<=column family \')\w+(?=\')", line)[0]
                testMap[sstFilename] = [sstLevel, cf]

        # Compare the map obtained from dump_live_files and the map
        # obtained from list_live_files_metadata. Everything should match.
        self.assertEqual(referenceMap, testMap)

    def getManifests(self, directory):
        return glob.glob(directory + "/MANIFEST-*")

    def getSSTFiles(self, directory):
        return glob.glob(directory + "/*.sst")

    def getWALFiles(self, directory):
        return glob.glob(directory + "/*.log")

    def getBlobFiles(self, directory):
        return glob.glob(directory + "/*.blob")

    def copyManifests(self, src, dest):
        return 0 == run_err_null("cp " + src + " " + dest)

    def testManifestDump(self):
        print("Running testManifestDump...")
        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put 1 1 --create_if_missing", "OK")
        self.assertRunOK("put 2 2", "OK")
        self.assertRunOK("put 3 3", "OK")
        # Pattern to expect from manifest_dump.
        num = "[0-9]+"
        st = ".*"
        subpat = st + " seq:" + num + ", type:" + num
        regex = num + ":" + num + r"\[" + subpat + ".." + subpat + r"\]"
        expected_pattern = re.compile(regex)
        cmd = "manifest_dump --db=%s"
        manifest_files = self.getManifests(dbPath)
        self.assertTrue(len(manifest_files) == 1)
        # Test with the default manifest file in dbPath.
        self.assertRunOKFull(
            cmd % dbPath, expected_pattern, unexpected=False, isPattern=True
        )
        self.copyManifests(manifest_files[0], manifest_files[0] + "1")
        manifest_files = self.getManifests(dbPath)
        self.assertTrue(len(manifest_files) == 2)
        # Test with multiple manifest files in dbPath.
        self.assertRunFAILFull(cmd % dbPath)
        # Running it with the copy we just created should pass.
        self.assertRunOKFull(
            (cmd + " --path=%s") % (dbPath, manifest_files[1]),
            expected_pattern,
            unexpected=False,
            isPattern=True,
        )
        # Make sure that using the dump with --path will result in identical
        # output as just using manifest_dump.
        cmd = "dump --path=%s"
        self.assertRunOKFull(
            (cmd) % (manifest_files[1]),
            expected_pattern,
            unexpected=False,
            isPattern=True,
        )

        # Check if null characters doesn't infer with output format.
        self.assertRunOK("put a1 b1", "OK")
        self.assertRunOK("put a2 b2", "OK")
        self.assertRunOK("put --hex 0x12000DA0 0x80C0000B", "OK")
        self.assertRunOK("put --hex 0x7200004f 0x80000004", "OK")
        self.assertRunOK("put --hex 0xa000000a 0xf000000f", "OK")
        self.assertRunOK("put a3 b3", "OK")
        self.assertRunOK("put a4 b4", "OK")

        # Verifies that all "levels" are printed out.
        # There should be 66 mentions of levels.
        expected_verbose_output = re.compile("matched")
        # Test manifest_dump verbose and verify that key 0x7200004f
        # is present. Note that we are forced to use grep here because
        # an output with a non-terminating null character in it isn't piped
        # correctly through the Python subprocess object.
        # Also note that 0x72=r and 0x4f=O, hence the regex \'r.{2}O\'
        # (we cannot use null character in the subprocess input either,
        # so we have to use '.{2}')
        cmd_verbose = (
            "manifest_dump --verbose --db=%s | grep -aq $''r.{2}O'' && echo 'matched' || echo 'not matched'"
            % dbPath
        )

        self.assertRunOKFull(
            cmd_verbose, expected_verbose_output, unexpected=False, isPattern=True
        )

    def testGetProperty(self):
        print("Running testGetProperty...")
        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put 1 1 --create_if_missing", "OK")
        self.assertRunOK("put 2 2", "OK")
        # A "string" property
        cmd = "--db=%s get_property rocksdb.estimate-num-keys"
        self.assertRunOKFull(cmd % dbPath, "rocksdb.estimate-num-keys: 2")
        # A "map" property
        # FIXME: why doesn't this pick up two entries?
        cmd = "--db=%s get_property rocksdb.aggregated-table-properties"
        part = "rocksdb.aggregated-table-properties.num_entries: "
        expected_pattern = re.compile(part)
        self.assertRunOKFull(
            cmd % dbPath, expected_pattern, unexpected=False, isPattern=True
        )
        # An invalid property
        cmd = "--db=%s get_property rocksdb.this-property-does-not-exist"
        self.assertRunFAILFull(cmd % dbPath)

    def testSSTDump(self):
        print("Running testSSTDump...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put sst1 sst1_val --create_if_missing", "OK")
        self.assertRunOK("put sst2 sst2_val --enable_blob_files", "OK")
        self.assertRunOK("get sst1", "sst1_val")

        # Pattern to expect from SST dump.
        regex = ".*Sst file format:.*\n.*\\[blob ref\\].*"
        expected_pattern = re.compile(regex)

        sst_files = self.getSSTFiles(dbPath)
        self.assertTrue(len(sst_files) >= 1)
        cmd = "dump --path=%s --decode_blob_index"
        self.assertRunOKFull(
            (cmd) % (sst_files[0]), expected_pattern, unexpected=False, isPattern=True
        )

    def testBlobDump(self):
        print("Running testBlobDump")
        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("batchput x1 y1 --create_if_missing --enable_blob_files", "OK")
        self.assertRunOK(
            'batchput --enable_blob_files x2 y2 x3 y3 "x4 abc" "y4 xyz"', "OK"
        )

        # Pattern to expect from blob file dump.
        regex = r".*Blob log header[\s\S]*Blob log footer[\s\S]*Read record[\s\S]*Summary"  # noqa
        expected_pattern = re.compile(regex)
        blob_files = self.getBlobFiles(dbPath)
        self.assertTrue(len(blob_files) >= 1)
        cmd = "dump --path=%s --dump_uncompressed_blobs"
        self.assertRunOKFull(
            (cmd) % (blob_files[0]), expected_pattern, unexpected=False, isPattern=True
        )

    def testWALDump(self):
        print("Running testWALDump...")

        dbPath = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put wal1 wal1_val --create_if_missing", "OK")
        self.assertRunOK("put wal2 wal2_val", "OK")
        self.assertRunOK("get wal1", "wal1_val")

        # Pattern to expect from WAL dump.
        regex = r"^Sequence,Count,ByteSize,Physical Offset,Key\(s\).*"
        expected_pattern = re.compile(regex)

        wal_files = self.getWALFiles(dbPath)
        self.assertTrue(len(wal_files) >= 1)
        cmd = "dump --path=%s"
        self.assertRunOKFull(
            (cmd) % (wal_files[0]), expected_pattern, unexpected=False, isPattern=True
        )

    def testListColumnFamilies(self):
        print("Running testListColumnFamilies...")
        self.assertRunOK("put x1 y1 --create_if_missing", "OK")
        cmd = 'list_column_families | grep -v "Column families"'
        # Test on valid dbPath.
        self.assertRunOK(cmd, "{default}")
        # Test on empty path.
        self.assertRunFAIL(cmd)

    def testColumnFamilies(self):
        print("Running testColumnFamilies...")
        _ = os.path.join(self.TMP_DIR, self.DB_NAME)
        self.assertRunOK("put cf1_1 1 --create_if_missing", "OK")
        self.assertRunOK("put cf1_2 2 --create_if_missing", "OK")
        self.assertRunOK("put cf1_3 3 --try_load_options", "OK")
        # Given non-default column family to single CF DB.
        self.assertRunFAIL("get cf1_1 --column_family=two")
        self.assertRunOK("create_column_family two", "OK")
        self.assertRunOK("put cf2_1 1 --create_if_missing --column_family=two", "OK")
        self.assertRunOK("put cf2_2 2 --create_if_missing --column_family=two", "OK")
        self.assertRunOK("delete cf1_2", "OK")
        self.assertRunOK("create_column_family three", "OK")
        self.assertRunOK("delete cf2_2 --column_family=two", "OK")
        self.assertRunOK("put cf3_1 3 --create_if_missing --column_family=three", "OK")
        self.assertRunOK("get cf1_1 --column_family=default", "1")
        self.assertRunOK("dump --column_family=two", "cf2_1 ==> 1\nKeys in range: 1")
        self.assertRunOK(
            "dump --column_family=two --try_load_options",
            "cf2_1 ==> 1\nKeys in range: 1",
        )
        self.assertRunOK("dump", "cf1_1 ==> 1\ncf1_3 ==> 3\nKeys in range: 2")
        self.assertRunOK("get cf2_1 --column_family=two", "1")
        self.assertRunOK("get cf3_1 --column_family=three", "3")
        self.assertRunOK("drop_column_family three", "OK")
        # non-existing column family.
        self.assertRunFAIL("get cf3_1 --column_family=four")
        self.assertRunFAIL("drop_column_family four")

    def testIngestExternalSst(self):
        print("Running testIngestExternalSst...")

        # Dump, load, write external sst and ingest it in another db
        dbPath = os.path.join(self.TMP_DIR, "db1")
        self.assertRunOK(
            "batchput --db=%s --create_if_missing x1 y1 x2 y2 x3 y3 x4 y4" % dbPath,
            "OK",
        )
        self.assertRunOK(
            "scan --db=%s" % dbPath, "x1 ==> y1\nx2 ==> y2\nx3 ==> y3\nx4 ==> y4"
        )
        dumpFilePath = os.path.join(self.TMP_DIR, "dump1")
        with open(dumpFilePath, "w") as f:
            f.write("x1 ==> y10\nx2 ==> y20\nx3 ==> y30\nx4 ==> y40")
        externSstPath = os.path.join(self.TMP_DIR, "extern_data1.sst")
        self.assertTrue(
            self.writeExternSst(
                "--create_if_missing --db=%s" % dbPath, dumpFilePath, externSstPath
            )
        )
        # cannot ingest if allow_global_seqno is false
        self.assertFalse(
            self.ingestExternSst(
                "--create_if_missing --allow_global_seqno=false --db=%s" % dbPath,
                externSstPath,
            )
        )
        self.assertTrue(
            self.ingestExternSst(
                "--create_if_missing --allow_global_seqno --db=%s" % dbPath,
                externSstPath,
            )
        )
        self.assertRunOKFull(
            "scan --db=%s" % dbPath, "x1 ==> y10\nx2 ==> y20\nx3 ==> y30\nx4 ==> y40"
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import copy
import random

from advisor.db_log_parser import NO_COL_FAMILY
from advisor.db_options_parser import DatabaseOptions
from advisor.rule_parser import Suggestion


class ConfigOptimizer:
    SCOPE = "scope"
    SUGG_VAL = "suggested values"

    @staticmethod
    def apply_action_on_value(old_value, action, suggested_values):
        chosen_sugg_val = None
        if suggested_values:
            chosen_sugg_val = random.choice(list(suggested_values))
        new_value = None
        if action is Suggestion.Action.set or not old_value:
            assert chosen_sugg_val
            new_value = chosen_sugg_val
        else:
            # For increase/decrease actions, currently the code tries to make
            # a 30% change in the option's value per iteration. An addend is
            # also present (+1 or -1) to handle the cases when the option's
            # old value was 0 or the final int() conversion suppressed the 30%
            # change made to the option
            old_value = float(old_value)
            mul = 0
            add = 0
            if action is Suggestion.Action.increase:
                if old_value < 0:
                    mul = 0.7
                    add = 2
                else:
                    mul = 1.3
                    add = 2
            elif action is Suggestion.Action.decrease:
                if old_value < 0:
                    mul = 1.3
                    add = -2
                else:
                    mul = 0.7
                    add = -2
            new_value = int(old_value * mul + add)
        return new_value

    @staticmethod
    def improve_db_config(options, rule, suggestions_dict):
        # this method takes ONE 'rule' and applies all its suggestions on the
        # appropriate options
        required_options = []
        rule_suggestions = []
        for sugg_name in rule.get_suggestions():
            option = suggestions_dict[sugg_name].option
            action = suggestions_dict[sugg_name].action
            # A Suggestion in the rules spec must have the 'option' and
            # 'action' fields defined, always call perform_checks() method
            # after parsing the rules file using RulesSpec
            assert option
            assert action
            required_options.append(option)
            rule_suggestions.append(suggestions_dict[sugg_name])
        current_config = options.get_options(required_options)
        # Create the updated configuration from the rule's suggestions
        updated_config = {}
        for sugg in rule_suggestions:
            # case: when the option is not present in the current configuration
            if sugg.option not in current_config:
                try:
                    new_value = ConfigOptimizer.apply_action_on_value(
                        None, sugg.action, sugg.suggested_values
                    )
                    if sugg.option not in updated_config:
                        updated_config[sugg.option] = {}
                    if DatabaseOptions.is_misc_option(sugg.option):
                        # this suggestion is on an option that is not yet
                        # supported by the Rocksdb OPTIONS file and so it is
                        # not prefixed by a section type.
                        updated_config[sugg.option][NO_COL_FAMILY] = new_value
                    else:
                        for col_fam in rule.get_trigger_column_families():
                            updated_config[sugg.option][col_fam] = new_value
                except AssertionError:
                    print(
                        "WARNING(ConfigOptimizer): provide suggested_values "
                        + "for "
                        + sugg.option
                    )
                continue
            # case: when the option is present in the current configuration
            if NO_COL_FAMILY in current_config[sugg.option]:
                old_value = current_config[sugg.option][NO_COL_FAMILY]
                try:
                    new_value = ConfigOptimizer.apply_action_on_value(
                        old_value, sugg.action, sugg.suggested_values
                    )
                    if sugg.option not in updated_config:
                        updated_config[sugg.option] = {}
                    updated_config[sugg.option][NO_COL_FAMILY] = new_value
                except AssertionError:
                    print(
                        "WARNING(ConfigOptimizer): provide suggested_values "
                        + "for "
                        + sugg.option
                    )
            else:
                for col_fam in rule.get_trigger_column_families():
                    old_value = None
                    if col_fam in current_config[sugg.option]:
                        old_value = current_config[sugg.option][col_fam]
                    try:
                        new_value = ConfigOptimizer.apply_action_on_value(
                            old_value, sugg.action, sugg.suggested_values
                        )
                        if sugg.option not in updated_config:
                            updated_config[sugg.option] = {}
                        updated_config[sugg.option][col_fam] = new_value
                    except AssertionError:
                        print(
                            "WARNING(ConfigOptimizer): provide "
                            + "suggested_values for "
                            + sugg.option
                        )
        return current_config, updated_config

    @staticmethod
    def pick_rule_to_apply(rules, last_rule_name, rules_tried, backtrack):
        if not rules:
            print("\nNo more rules triggered!")
            return None
        # if the last rule provided an improvement in the database performance,
        # and it was triggered again (i.e. it is present in 'rules'), then pick
        # the same rule for this iteration too.
        if last_rule_name and not backtrack:
            for rule in rules:
                if rule.name == last_rule_name:
                    return rule
        # there was no previous rule OR the previous rule did not improve db
        # performance OR it was not triggered for this iteration,
        # then pick another rule that has not been tried yet
        for rule in rules:
            if rule.name not in rules_tried:
                return rule
        print("\nAll rules have been exhausted")
        return None

    @staticmethod
    def apply_suggestions(
        triggered_rules,
        current_rule_name,
        rules_tried,
        backtrack,
        curr_options,
        suggestions_dict,
    ):
        curr_rule = ConfigOptimizer.pick_rule_to_apply(
            triggered_rules, current_rule_name, rules_tried, backtrack
        )
        if not curr_rule:
            return tuple([None] * 4)
        # if a rule has been picked for improving db_config, update rules_tried
        rules_tried.add(curr_rule.name)
        # get updated config based on the picked rule
        curr_conf, updated_conf = ConfigOptimizer.improve_db_config(
            curr_options, curr_rule, suggestions_dict
        )
        conf_diff = DatabaseOptions.get_options_diff(curr_conf, updated_conf)
        if not conf_diff:  # the current and updated configs are the same
            (
                curr_rule,
                rules_tried,
                curr_conf,
                updated_conf,
            ) = ConfigOptimizer.apply_suggestions(
                triggered_rules,
                None,
                rules_tried,
                backtrack,
                curr_options,
                suggestions_dict,
            )
        print("returning from apply_suggestions")
        return (curr_rule, rules_tried, curr_conf, updated_conf)

    # TODO(poojam23): check if this method is required or can we directly set
    # the config equal to the curr_config
    @staticmethod
    def get_backtrack_config(curr_config, updated_config):
        diff = DatabaseOptions.get_options_diff(curr_config, updated_config)
        bt_config = {}
        for option in diff:
            bt_config[option] = {}
            for col_fam in diff[option]:
                bt_config[option][col_fam] = diff[option][col_fam][0]
        print(bt_config)
        return bt_config

    def __init__(self, bench_runner, db_options, rule_parser, base_db):
        self.bench_runner = bench_runner
        self.db_options = db_options
        self.rule_parser = rule_parser
        self.base_db_path = base_db

    def run(self):
        # In every iteration of this method's optimization loop we pick ONE
        # RULE from all the triggered rules and apply all its suggestions to
        # the appropriate options.
        # bootstrapping the optimizer
        print("Bootstrapping optimizer:")
        options = copy.deepcopy(self.db_options)
        old_data_sources, old_metric = self.bench_runner.run_experiment(
            options, self.base_db_path
        )
        print("Initial metric: " + str(old_metric))
        self.rule_parser.load_rules_from_spec()
        self.rule_parser.perform_section_checks()
        triggered_rules = self.rule_parser.get_triggered_rules(
            old_data_sources, options.get_column_families()
        )
        print("\nTriggered:")
        self.rule_parser.print_rules(triggered_rules)
        backtrack = False
        rules_tried = set()
        (
            curr_rule,
            rules_tried,
            curr_conf,
            updated_conf,
        ) = ConfigOptimizer.apply_suggestions(
            triggered_rules,
            None,
            rules_tried,
            backtrack,
            options,
            self.rule_parser.get_suggestions_dict(),
        )
        # the optimizer loop
        while curr_rule:
            print("\nRule picked for next iteration:")
            print(curr_rule.name)
            print("\ncurrent config:")
            print(curr_conf)
            print("updated config:")
            print(updated_conf)
            options.update_options(updated_conf)
            # run bench_runner with updated config
            new_data_sources, new_metric = self.bench_runner.run_experiment(
                options, self.base_db_path
            )
            print("\nnew metric: " + str(new_metric))
            backtrack = not self.bench_runner.is_metric_better(new_metric, old_metric)
            # update triggered_rules, metric, data_sources, if required
            if backtrack:
                # revert changes to options config
                print("\nBacktracking to previous configuration")
                backtrack_conf = ConfigOptimizer.get_backtrack_config(
                    curr_conf, updated_conf
                )
                options.update_options(backtrack_conf)
            else:
                # run advisor on new data sources
                self.rule_parser.load_rules_from_spec()  # reboot the advisor
                self.rule_parser.perform_section_checks()
                triggered_rules = self.rule_parser.get_triggered_rules(
                    new_data_sources, options.get_column_families()
                )
                print("\nTriggered:")
                self.rule_parser.print_rules(triggered_rules)
                old_metric = new_metric
                old_data_sources = new_data_sources
                rules_tried = set()
            # pick rule to work on and set curr_rule to that
            (
                curr_rule,
                rules_tried,
                curr_conf,
                updated_conf,
            ) = ConfigOptimizer.apply_suggestions(
                triggered_rules,
                curr_rule.name,
                rules_tried,
                backtrack,
                options,
                self.rule_parser.get_suggestions_dict(),
            )
        # return the final database options configuration
        return options



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import os
import random
import sys

from block_cache_pysim import (
    ARCCache,
    CacheEntry,
    create_cache,
    GDSizeCache,
    HashTable,
    HyperbolicPolicy,
    kMicrosInSecond,
    kSampleSize,
    LFUPolicy,
    LinUCBCache,
    LRUCache,
    LRUPolicy,
    MRUPolicy,
    OPTCache,
    OPTCacheEntry,
    run,
    ThompsonSamplingCache,
    TraceCache,
    TraceRecord,
)


def test_hash_table():
    print("Test hash table")
    table = HashTable()
    data_size = 10000
    for i in range(data_size):
        table.insert(f"k{i}", i, f"v{i}")
    for i in range(data_size):
        assert table.lookup(f"k{i}", i) is not None
    for i in range(data_size):
        table.delete(f"k{i}", i)
    for i in range(data_size):
        assert table.lookup(f"k{i}", i) is None

    truth_map = {}
    n = 1000000
    records = 100
    for i in range(n):
        key_id = random.randint(0, records)
        v = random.randint(0, records)
        key = f"k{key_id}"
        value = CacheEntry(v, v, v, v, v, v, v)
        action = random.randint(0, 10)
        assert len(truth_map) == table.elements, "{} {} {}".format(
            len(truth_map), table.elements, i
        )
        if action <= 8:
            if key in truth_map:
                assert table.lookup(key, key_id) is not None
                assert truth_map[key].value_size == table.lookup(key, key_id).value_size
            else:
                assert table.lookup(key, key_id) is None
            table.insert(key, key_id, value)
            truth_map[key] = value
        else:
            deleted = table.delete(key, key_id)
            if deleted:
                assert key in truth_map
            if key in truth_map:
                del truth_map[key]

    # Check all keys are unique in the sample set.
    for _i in range(10):
        samples = table.random_sample(kSampleSize)
        unique_keys = {}
        for sample in samples:
            unique_keys[sample.key] = True
        assert len(samples) == len(unique_keys)

    assert len(table) == len(truth_map)
    for key in truth_map:
        assert table.lookup(key, int(key[1:])) is not None
        assert truth_map[key].value_size == table.lookup(key, int(key[1:])).value_size
    print("Test hash table: Success")


def assert_metrics(cache, expected_value, expected_value_size=1, custom_hashtable=True):
    assert cache.used_size == expected_value[0], "Expected {}, Actual {}".format(
        expected_value[0], cache.used_size
    )
    assert (
        cache.miss_ratio_stats.num_accesses == expected_value[1]
    ), "Expected {}, Actual {}".format(
        expected_value[1], cache.miss_ratio_stats.num_accesses
    )
    assert (
        cache.miss_ratio_stats.num_misses == expected_value[2]
    ), "Expected {}, Actual {}".format(
        expected_value[2], cache.miss_ratio_stats.num_misses
    )
    assert len(cache.table) == len(expected_value[3]) + len(
        expected_value[4]
    ), "Expected {}, Actual {}".format(
        len(expected_value[3]) + len(expected_value[4]), cache.table.elements
    )
    for expeceted_k in expected_value[3]:
        if custom_hashtable:
            val = cache.table.lookup(f"b{expeceted_k}", expeceted_k)
        else:
            val = cache.table[f"b{expeceted_k}"]
        assert val is not None, "Expected {} Actual: Not Exist {}, Table: {}".format(
            expeceted_k, expected_value, cache.table
        )
        assert val.value_size == expected_value_size
    for expeceted_k in expected_value[4]:
        if custom_hashtable:
            val = cache.table.lookup(f"g0-{expeceted_k}", expeceted_k)
        else:
            val = cache.table[f"g0-{expeceted_k}"]
        assert val is not None
        assert val.value_size == expected_value_size


# Access k1, k1, k2, k3, k3, k3, k4
# When k4 is inserted,
#   LRU should evict k1.
#   LFU should evict k2.
#   MRU should evict k3.
def test_cache(cache, expected_value, custom_hashtable=True):
    k1 = TraceRecord(
        access_time=0,
        block_id=1,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,
        key_id=1,
        kv_size=5,
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=0,
    )
    k2 = TraceRecord(
        access_time=1,
        block_id=2,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,
        key_id=1,
        kv_size=5,
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=0,
    )
    k3 = TraceRecord(
        access_time=2,
        block_id=3,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,
        key_id=1,
        kv_size=5,
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=0,
    )
    k4 = TraceRecord(
        access_time=3,
        block_id=4,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,
        key_id=1,
        kv_size=5,
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=0,
    )
    sequence = [k1, k1, k2, k3, k3, k3]
    index = 0
    expected_values = []
    # Access k1, miss.
    expected_values.append([1, 1, 1, [1], []])
    # Access k1, hit.
    expected_values.append([1, 2, 1, [1], []])
    # Access k2, miss.
    expected_values.append([2, 3, 2, [1, 2], []])
    # Access k3, miss.
    expected_values.append([3, 4, 3, [1, 2, 3], []])
    # Access k3, hit.
    expected_values.append([3, 5, 3, [1, 2, 3], []])
    # Access k3, hit.
    expected_values.append([3, 6, 3, [1, 2, 3], []])
    access_time = 0
    for access in sequence:
        access.access_time = access_time
        cache.access(access)
        assert_metrics(
            cache,
            expected_values[index],
            expected_value_size=1,
            custom_hashtable=custom_hashtable,
        )
        access_time += 1
        index += 1
    k4.access_time = access_time
    cache.access(k4)
    assert_metrics(
        cache, expected_value, expected_value_size=1, custom_hashtable=custom_hashtable
    )


def test_lru_cache(cache, custom_hashtable):
    print("Test LRU cache")
    # Access k4, miss. evict k1
    test_cache(cache, [3, 7, 4, [2, 3, 4], []], custom_hashtable)
    print("Test LRU cache: Success")


def test_mru_cache():
    print("Test MRU cache")
    policies = []
    policies.append(MRUPolicy())
    # Access k4, miss. evict k3
    test_cache(
        ThompsonSamplingCache(3, False, policies, cost_class_label=None),
        [3, 7, 4, [1, 2, 4], []],
    )
    print("Test MRU cache: Success")


def test_lfu_cache():
    print("Test LFU cache")
    policies = []
    policies.append(LFUPolicy())
    # Access k4, miss. evict k2
    test_cache(
        ThompsonSamplingCache(3, False, policies, cost_class_label=None),
        [3, 7, 4, [1, 3, 4], []],
    )
    print("Test LFU cache: Success")


def test_mix(cache):
    print(f"Test Mix {cache.cache_name()} cache")
    n = 100000
    records = 100
    block_size_table = {}
    trace_num_misses = 0
    for i in range(n):
        key_id = random.randint(0, records)
        vs = random.randint(0, 10)
        now = i * kMicrosInSecond
        block_size = vs
        if key_id in block_size_table:
            block_size = block_size_table[key_id]
        else:
            block_size_table[key_id] = block_size
        is_hit = key_id % 2
        if is_hit == 0:
            trace_num_misses += 1
        k = TraceRecord(
            access_time=now,
            block_id=key_id,
            block_type=1,
            block_size=block_size,
            cf_id=0,
            cf_name="",
            level=0,
            fd=0,
            caller=1,
            no_insert=0,
            get_id=key_id,
            key_id=key_id,
            kv_size=5,
            is_hit=is_hit,
            referenced_key_exist_in_block=1,
            num_keys_in_block=0,
            table_id=0,
            seq_number=0,
            block_key_size=0,
            key_size=0,
            block_offset_in_file=0,
            next_access_seq_no=vs,
        )
        cache.access(k)
    assert cache.miss_ratio_stats.miss_ratio() > 0
    if cache.cache_name() == "Trace":
        assert cache.miss_ratio_stats.num_accesses == n
        assert cache.miss_ratio_stats.num_misses == trace_num_misses
    else:
        assert cache.used_size <= cache.cache_size
        all_values = cache.table.values()
        cached_size = 0
        for value in all_values:
            cached_size += value.value_size
        assert cached_size == cache.used_size, "Expeced {} Actual {}".format(
            cache.used_size, cached_size
        )
    print(f"Test Mix {cache.cache_name()} cache: Success")


def test_end_to_end():
    print("Test All caches")
    n = 100000
    nblocks = 1000
    block_size = 16 * 1024
    ncfs = 7
    nlevels = 6
    nfds = 100000
    trace_file_path = "test_trace"
    # All blocks are of the same size so that OPT must achieve the lowest miss
    # ratio.
    with open(trace_file_path, "w+") as trace_file:
        access_records = ""
        for i in range(n):
            key_id = random.randint(0, nblocks)
            cf_id = random.randint(0, ncfs)
            level = random.randint(0, nlevels)
            fd = random.randint(0, nfds)
            now = i * kMicrosInSecond
            access_record = ""
            access_record += f"{now},"
            access_record += f"{key_id},"
            access_record += f"{9},"  # block type
            access_record += f"{block_size},"  # block size
            access_record += f"{cf_id},"
            access_record += f"cf_{cf_id},"
            access_record += f"{level},"
            access_record += f"{fd},"
            access_record += f"{key_id % 3},"  # caller
            access_record += f"{0},"  # no insert
            access_record += f"{i},"  # get_id
            access_record += f"{i},"  # key_id
            access_record += f"{100},"  # kv_size
            access_record += f"{1},"  # is_hit
            access_record += f"{1},"  # referenced_key_exist_in_block
            access_record += f"{10},"  # num_keys_in_block
            access_record += f"{1},"  # table_id
            access_record += f"{0},"  # seq_number
            access_record += f"{10},"  # block key size
            access_record += f"{20},"  # key size
            access_record += f"{0},"  # block offset
            access_record = access_record[:-1]
            access_records += access_record + "\n"
        trace_file.write(access_records)

    print("Test All caches: Start testing caches")
    cache_size = block_size * nblocks / 10
    downsample_size = 1
    cache_ms = {}
    for cache_type in [
        "ts",
        "opt",
        "lru",
        "pylru",
        "linucb",
        "gdsize",
        "pyccbt",
        "pycctbbt",
    ]:
        cache = create_cache(cache_type, cache_size, downsample_size)
        run(trace_file_path, cache_type, cache, 0, -1, "all")
        cache_ms[cache_type] = cache
        assert cache.miss_ratio_stats.num_accesses == n

    for cache_type in cache_ms:
        cache = cache_ms[cache_type]
        ms = cache.miss_ratio_stats.miss_ratio()
        assert ms <= 100.0 and ms >= 0.0
        # OPT should perform the best.
        assert cache_ms["opt"].miss_ratio_stats.miss_ratio() <= ms
        assert cache.used_size <= cache.cache_size
        all_values = cache.table.values()
        cached_size = 0
        for value in all_values:
            cached_size += value.value_size
        assert cached_size == cache.used_size, "Expeced {} Actual {}".format(
            cache.used_size, cached_size
        )
        print(f"Test All {cache.cache_name()}: Success")

    os.remove(trace_file_path)
    print("Test All: Success")


def test_hybrid(cache):
    print(f"Test {cache.cache_name()} cache")
    k = TraceRecord(
        access_time=0,
        block_id=1,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,  # the first get request.
        key_id=1,
        kv_size=0,  # no size.
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=0,
    )
    cache.access(k)  # Expect a miss.
    # used size, num accesses, num misses, hash table size, blocks, get keys.
    assert_metrics(cache, [1, 1, 1, [1], []])
    k.access_time += 1
    k.kv_size = 1
    k.block_id = 2
    cache.access(k)  # k should be inserted.
    assert_metrics(cache, [3, 2, 2, [1, 2], [1]])
    k.access_time += 1
    k.block_id = 3
    cache.access(k)  # k should not be inserted again.
    assert_metrics(cache, [4, 3, 3, [1, 2, 3], [1]])
    # A second get request referencing the same key.
    k.access_time += 1
    k.get_id = 2
    k.block_id = 4
    k.kv_size = 0
    cache.access(k)  # k should observe a hit. No block access.
    assert_metrics(cache, [4, 4, 3, [1, 2, 3], [1]])

    # A third get request searches three files, three different keys.
    # And the second key observes a hit.
    k.access_time += 1
    k.kv_size = 1
    k.get_id = 3
    k.block_id = 3
    k.key_id = 2
    cache.access(k)  # k should observe a miss. block 3 observes a hit.
    assert_metrics(cache, [5, 5, 3, [1, 2, 3], [1, 2]])

    k.access_time += 1
    k.kv_size = 1
    k.get_id = 3
    k.block_id = 4
    k.kv_size = 1
    k.key_id = 1
    cache.access(k)  # k1 should observe a hit.
    assert_metrics(cache, [5, 6, 3, [1, 2, 3], [1, 2]])

    k.access_time += 1
    k.kv_size = 1
    k.get_id = 3
    k.block_id = 4
    k.kv_size = 1
    k.key_id = 3
    # k3 should observe a miss.
    # However, as the get already complete, we should not access k3 any more.
    cache.access(k)
    assert_metrics(cache, [5, 7, 3, [1, 2, 3], [1, 2]])

    # A fourth get request searches one file and two blocks. One row key.
    k.access_time += 1
    k.get_id = 4
    k.block_id = 5
    k.key_id = 4
    k.kv_size = 1
    cache.access(k)
    assert_metrics(cache, [7, 8, 4, [1, 2, 3, 5], [1, 2, 4]])

    # A bunch of insertions which evict cached row keys.
    for i in range(6, 100):
        k.access_time += 1
        k.get_id = 0
        k.block_id = i
        cache.access(k)

    k.get_id = 4
    k.block_id = 100  # A different block.
    k.key_id = 4  # Same row key and should not be inserted again.
    k.kv_size = 1
    cache.access(k)
    assert_metrics(
        cache, [kSampleSize, 103, 99, [i for i in range(101 - kSampleSize, 101)], []]
    )
    print(f"Test {cache.cache_name()} cache: Success")


def test_opt_cache():
    print("Test OPT cache")
    cache = OPTCache(3)
    # seq:         0,  1,  2,  3,  4,  5,  6,  7,  8
    # key:         k1, k2, k3, k4, k5, k6, k7, k1, k8
    # next_access: 7,  19, 18, M,  M,  17, 16, 25, M
    k = TraceRecord(
        access_time=0,
        block_id=1,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,  # the first get request.
        key_id=1,
        kv_size=0,  # no size.
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=7,
    )
    cache.access(k)
    assert_metrics(
        cache, [1, 1, 1, [1], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 2
    k.next_access_seq_no = 19
    cache.access(k)
    assert_metrics(
        cache, [2, 2, 2, [1, 2], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 3
    k.next_access_seq_no = 18
    cache.access(k)
    assert_metrics(
        cache, [3, 3, 3, [1, 2, 3], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 4
    k.next_access_seq_no = sys.maxsize  # Never accessed again.
    cache.access(k)
    # Evict 2 since its next access 19 is the furthest in the future.
    assert_metrics(
        cache, [3, 4, 4, [1, 3, 4], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 5
    k.next_access_seq_no = sys.maxsize  # Never accessed again.
    cache.access(k)
    # Evict 4 since its next access MAXINT is the furthest in the future.
    assert_metrics(
        cache, [3, 5, 5, [1, 3, 5], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 6
    k.next_access_seq_no = 17
    cache.access(k)
    # Evict 5 since its next access MAXINT is the furthest in the future.
    assert_metrics(
        cache, [3, 6, 6, [1, 3, 6], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 7
    k.next_access_seq_no = 16
    cache.access(k)
    # Evict 3 since its next access 18 is the furthest in the future.
    assert_metrics(
        cache, [3, 7, 7, [1, 6, 7], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 1
    k.next_access_seq_no = 25
    cache.access(k)
    assert_metrics(
        cache, [3, 8, 7, [1, 6, 7], []], expected_value_size=1, custom_hashtable=False
    )
    k.access_time += 1
    k.block_id = 8
    k.next_access_seq_no = sys.maxsize
    cache.access(k)
    # Evict 1 since its next access 25 is the furthest in the future.
    assert_metrics(
        cache, [3, 9, 8, [6, 7, 8], []], expected_value_size=1, custom_hashtable=False
    )

    # Insert a large kv pair to evict all keys.
    k.access_time += 1
    k.block_id = 10
    k.block_size = 3
    k.next_access_seq_no = sys.maxsize
    cache.access(k)
    assert_metrics(
        cache, [3, 10, 9, [10], []], expected_value_size=3, custom_hashtable=False
    )
    print("Test OPT cache: Success")


def test_trace_cache():
    print("Test trace cache")
    cache = TraceCache(0)
    k = TraceRecord(
        access_time=0,
        block_id=1,
        block_type=1,
        block_size=1,
        cf_id=0,
        cf_name="",
        level=0,
        fd=0,
        caller=1,
        no_insert=0,
        get_id=1,
        key_id=1,
        kv_size=0,
        is_hit=1,
        referenced_key_exist_in_block=1,
        num_keys_in_block=0,
        table_id=0,
        seq_number=0,
        block_key_size=0,
        key_size=0,
        block_offset_in_file=0,
        next_access_seq_no=7,
    )
    cache.access(k)
    assert cache.miss_ratio_stats.num_accesses == 1
    assert cache.miss_ratio_stats.num_misses == 0
    k.is_hit = 0
    cache.access(k)
    assert cache.miss_ratio_stats.num_accesses == 2
    assert cache.miss_ratio_stats.num_misses == 1
    print("Test trace cache: Success")


if __name__ == "__main__":
    test_hash_table()
    test_trace_cache()
    test_opt_cache()
    test_lru_cache(
        ThompsonSamplingCache(
            3, enable_cache_row_key=0, policies=[LRUPolicy()], cost_class_label=None
        ),
        custom_hashtable=True,
    )
    test_lru_cache(LRUCache(3, enable_cache_row_key=0), custom_hashtable=False)
    test_mru_cache()
    test_lfu_cache()
    test_hybrid(
        ThompsonSamplingCache(
            kSampleSize,
            enable_cache_row_key=1,
            policies=[LRUPolicy()],
            cost_class_label=None,
        )
    )
    test_hybrid(
        LinUCBCache(
            kSampleSize,
            enable_cache_row_key=1,
            policies=[LRUPolicy()],
            cost_class_label=None,
        )
    )
    for cache_type in [
        "ts",
        "opt",
        "arc",
        "pylfu",
        "pymru",
        "trace",
        "pyhb",
        "lru",
        "pylru",
        "linucb",
        "gdsize",
        "pycctbbt",
        "pycctb",
        "pyccbt",
    ]:
        for enable_row_cache in [0, 1, 2]:
            cache_type_str = cache_type
            if cache_type != "opt" and cache_type != "trace":
                if enable_row_cache == 1:
                    cache_type_str += "_hybrid"
                elif enable_row_cache == 2:
                    cache_type_str += "_hybridn"
            test_mix(create_cache(cache_type_str, cache_size=100, downsample_size=1))
    test_end_to_end()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

try:
    from builtins import object, str
except ImportError:
    from __builtin__ import object, str
import pprint

import targets_cfg

class LiteralValue:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

def smart_quote_value(val):
    if isinstance(val, LiteralValue):
        return str(val)
    return '"%s"' % val

def pretty_list(lst, indent=8):
    if lst is None or len(lst) == 0:
        return ""

    if len(lst) == 1:
        return smart_quote_value(lst[0])

    separator = ',\n%s' % (" " * indent)
    res = separator.join(sorted(map(smart_quote_value, lst)))
    res = "\n" + (" " * indent) + res + ',\n' + (" " * (indent - 4))
    return res


class TARGETSBuilder:
    def __init__(self, path, extra_argv):
        self.path = path
        header = targets_cfg.rocksdb_target_header_template.format(
            extra_argv=extra_argv
        )
        with open(path, "wb") as targets_file:
            targets_file.write(header.encode("utf-8"))
        self.total_lib = 0
        self.total_bin = 0
        self.total_test = 0
        self.tests_cfg = ""

    def add_library(
        self,
        name,
        srcs,
        deps=None,
        headers=None,
        extra_external_deps="",
        link_whole=False,
        external_dependencies=None,
        extra_test_libs=False,
    ):
        if headers is not None:
            if isinstance(headers, LiteralValue):
                headers = str(headers)
            else:
                headers = "[" + pretty_list(headers) + "]"
        else:
            headers = "[]"
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                targets_cfg.library_template.format(
                    name=name,
                    srcs=pretty_list(srcs),
                    headers=headers,
                    deps=pretty_list(deps),
                    extra_external_deps=extra_external_deps,
                    link_whole=link_whole,
                    external_dependencies=pretty_list(external_dependencies),
                    extra_test_libs=extra_test_libs,
                ).encode("utf-8")
            )
        self.total_lib = self.total_lib + 1

    def add_rocksdb_library(self, name, srcs, headers=None, external_dependencies=None):
        headers = "[" + pretty_list(headers) + "]"
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                targets_cfg.rocksdb_library_template.format(
                    name=name,
                    srcs=pretty_list(srcs),
                    headers=headers,
                    external_dependencies=pretty_list(external_dependencies),
                ).encode("utf-8")
            )
        self.total_lib = self.total_lib + 1

    def add_binary(
        self,
        name,
        srcs,
        deps=None,
        extra_preprocessor_flags=None,
        extra_bench_libs=False,
    ):
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                targets_cfg.binary_template.format(
                    name=name,
                    srcs=pretty_list(srcs),
                    deps=pretty_list(deps),
                    extra_preprocessor_flags=pretty_list(extra_preprocessor_flags),
                    extra_bench_libs=extra_bench_libs,
                ).encode("utf-8")
            )
        self.total_bin = self.total_bin + 1

    def add_c_test(self):
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                b"""
add_c_test_wrapper()
"""
            )

    def add_test_header(self):
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                b"""
        # Generate a test rule for each entry in ROCKS_TESTS
        # Do not build the tests in opt mode, since SyncPoint and other test code
        # will not be included.
"""
            )

    def add_fancy_bench_config(
        self,
        name,
        bench_config,
        slow,
        expected_runtime,
        sl_iterations,
        regression_threshold,
    ):
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                targets_cfg.fancy_bench_template.format(
                    name=name,
                    bench_config=pprint.pformat(bench_config),
                    slow=slow,
                    expected_runtime=expected_runtime,
                    sl_iterations=sl_iterations,
                    regression_threshold=regression_threshold,
                ).encode("utf-8")
            )

    def register_test(self, test_name, src, deps, extra_compiler_flags):
        with open(self.path, "ab") as targets_file:
            targets_file.write(
                targets_cfg.unittests_template.format(
                    test_name=test_name,
                    test_cc=str(src),
                    deps=deps,
                    extra_compiler_flags=extra_compiler_flags,
                ).encode("utf-8")
            )
        self.total_test = self.total_test + 1

    def export_file(self, name):
        with open(self.path, "a") as targets_file:
            targets_file.write(
                targets_cfg.export_file_template.format(name=name)
            )



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import os
import unittest

from advisor.db_bench_runner import DBBenchRunner
from advisor.db_log_parser import DataSource, NO_COL_FAMILY
from advisor.db_options_parser import DatabaseOptions


class TestDBBenchRunnerMethods(unittest.TestCase):
    def setUp(self):
        self.pos_args = [
            "./../../db_bench",
            "overwrite",
            "use_existing_db=true",
            "duration=10",
        ]
        self.bench_runner = DBBenchRunner(self.pos_args)
        this_path = os.path.abspath(os.path.dirname(__file__))
        options_path = os.path.join(this_path, "input_files/OPTIONS-000005")
        self.db_options = DatabaseOptions(options_path)

    def test_setup(self):
        self.assertEqual(self.bench_runner.db_bench_binary, self.pos_args[0])
        self.assertEqual(self.bench_runner.benchmark, self.pos_args[1])
        self.assertSetEqual(
            set(self.bench_runner.db_bench_args), set(self.pos_args[2:])
        )

    def test_get_info_log_file_name(self):
        log_file_name = DBBenchRunner.get_info_log_file_name(None, "random_path")
        self.assertEqual(log_file_name, "LOG")

        log_file_name = DBBenchRunner.get_info_log_file_name(
            "/dev/shm/", "/tmp/rocksdbtest-155919/dbbench/"
        )
        self.assertEqual(log_file_name, "tmp_rocksdbtest-155919_dbbench_LOG")

    def test_get_opt_args_str(self):
        misc_opt_dict = {"bloom_bits": 2, "empty_opt": None, "rate_limiter": 3}
        optional_args_str = DBBenchRunner.get_opt_args_str(misc_opt_dict)
        self.assertEqual(optional_args_str, " --bloom_bits=2 --rate_limiter=3")

    def test_get_log_options(self):
        db_path = "/tmp/rocksdb-155919/dbbench"
        # when db_log_dir is present in the db_options
        update_dict = {
            "DBOptions.db_log_dir": {NO_COL_FAMILY: "/dev/shm"},
            "DBOptions.stats_dump_period_sec": {NO_COL_FAMILY: "20"},
        }
        self.db_options.update_options(update_dict)
        log_file_prefix, stats_freq = self.bench_runner.get_log_options(
            self.db_options, db_path
        )
        self.assertEqual(log_file_prefix, "/dev/shm/tmp_rocksdb-155919_dbbench_LOG")
        self.assertEqual(stats_freq, 20)

        update_dict = {
            "DBOptions.db_log_dir": {NO_COL_FAMILY: None},
            "DBOptions.stats_dump_period_sec": {NO_COL_FAMILY: "30"},
        }
        self.db_options.update_options(update_dict)
        log_file_prefix, stats_freq = self.bench_runner.get_log_options(
            self.db_options, db_path
        )
        self.assertEqual(log_file_prefix, "/tmp/rocksdb-155919/dbbench/LOG")
        self.assertEqual(stats_freq, 30)

    def test_build_experiment_command(self):
        # add some misc_options to db_options
        update_dict = {
            "bloom_bits": {NO_COL_FAMILY: 2},
            "rate_limiter_bytes_per_sec": {NO_COL_FAMILY: 128000000},
        }
        self.db_options.update_options(update_dict)
        db_path = "/dev/shm"
        experiment_command = self.bench_runner._build_experiment_command(
            self.db_options, db_path
        )
        opt_args_str = DBBenchRunner.get_opt_args_str(
            self.db_options.get_misc_options()
        )
        opt_args_str += " --options_file=" + self.db_options.generate_options_config(
            "12345"
        )
        for arg in self.pos_args[2:]:
            opt_args_str += " --" + arg
        expected_command = (
            self.pos_args[0]
            + " --benchmarks="
            + self.pos_args[1]
            + " --statistics --perf_level=3 --db="
            + db_path
            + opt_args_str
        )
        self.assertEqual(experiment_command, expected_command)


class TestDBBenchRunner(unittest.TestCase):
    def setUp(self):
        # Note: the db_bench binary should be present in the rocksdb/ directory
        self.pos_args = [
            "./../../db_bench",
            "overwrite",
            "use_existing_db=true",
            "duration=20",
        ]
        self.bench_runner = DBBenchRunner(self.pos_args)
        this_path = os.path.abspath(os.path.dirname(__file__))
        options_path = os.path.join(this_path, "input_files/OPTIONS-000005")
        self.db_options = DatabaseOptions(options_path)

    def test_experiment_output(self):
        update_dict = {"bloom_bits": {NO_COL_FAMILY: 2}}
        self.db_options.update_options(update_dict)
        db_path = "/dev/shm"
        data_sources, throughput = self.bench_runner.run_experiment(
            self.db_options, db_path
        )
        self.assertEqual(
            data_sources[DataSource.Type.DB_OPTIONS][0].type, DataSource.Type.DB_OPTIONS
        )
        self.assertEqual(data_sources[DataSource.Type.LOG][0].type, DataSource.Type.LOG)
        self.assertEqual(len(data_sources[DataSource.Type.TIME_SERIES]), 2)
        self.assertEqual(
            data_sources[DataSource.Type.TIME_SERIES][0].type,
            DataSource.Type.TIME_SERIES,
        )
        self.assertEqual(
            data_sources[DataSource.Type.TIME_SERIES][1].type,
            DataSource.Type.TIME_SERIES,
        )
        self.assertEqual(data_sources[DataSource.Type.TIME_SERIES][1].stats_freq_sec, 0)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import shutil
import subprocess
import time

from advisor.bench_runner import BenchmarkRunner
from advisor.db_log_parser import DatabaseLogs, DataSource, NO_COL_FAMILY
from advisor.db_stats_fetcher import (
    DatabasePerfContext,
    LogStatsParser,
    OdsStatsFetcher,
)


"""
NOTE: This is not thread-safe, because the output file is simply overwritten.
"""


class DBBenchRunner(BenchmarkRunner):
    OUTPUT_FILE = "temp/dbbench_out.tmp"
    ERROR_FILE = "temp/dbbench_err.tmp"
    DB_PATH = "DB path"
    THROUGHPUT = "ops/sec"
    PERF_CON = " PERF_CONTEXT:"

    @staticmethod
    def is_metric_better(new_metric, old_metric):
        # for db_bench 'throughput' is the metric returned by run_experiment
        return new_metric >= old_metric

    @staticmethod
    def get_opt_args_str(misc_options_dict):
        # given a dictionary of options and their values, return a string
        # that can be appended as command-line arguments
        optional_args_str = ""
        for option_name, option_value in misc_options_dict.items():
            if option_value:
                optional_args_str += " --" + option_name + "=" + str(option_value)
        return optional_args_str

    def __init__(self, positional_args, ods_args=None):
        # parse positional_args list appropriately
        self.db_bench_binary = positional_args[0]
        self.benchmark = positional_args[1]
        self.db_bench_args = None
        if len(positional_args) > 2:
            # options list with each option given as "<option>=<value>"
            self.db_bench_args = positional_args[2:]
        # save ods_args, if provided
        self.ods_args = ods_args

    def _parse_output(self, get_perf_context=False):
        """
        Sample db_bench output after running 'readwhilewriting' benchmark:
        DB path: [/tmp/rocksdbtest-155919/dbbench]\n
        readwhilewriting : 16.582 micros/op 60305 ops/sec; 4.2 MB/s (3433828\
        of 5427999 found)\n
        PERF_CONTEXT:\n
        user_key_comparison_count = 500466712, block_cache_hit_count = ...\n
        """
        output = {self.THROUGHPUT: None, self.DB_PATH: None, self.PERF_CON: None}
        perf_context_begins = False
        with open(self.OUTPUT_FILE) as fp:
            for line in fp:
                if line.startswith(self.benchmark):
                    # line from sample output:
                    # readwhilewriting : 16.582 micros/op 60305 ops/sec; \
                    # 4.2 MB/s (3433828 of 5427999 found)\n
                    print(line)  # print output of the benchmark run
                    token_list = line.strip().split()
                    for ix, token in enumerate(token_list):
                        if token.startswith(self.THROUGHPUT):
                            # in above example, throughput = 60305 ops/sec
                            output[self.THROUGHPUT] = float(token_list[ix - 1])
                            break
                elif get_perf_context and line.startswith(self.PERF_CON):
                    # the following lines in the output contain perf context
                    # statistics (refer example above)
                    perf_context_begins = True
                elif get_perf_context and perf_context_begins:
                    # Sample perf_context output:
                    # user_key_comparison_count = 500, block_cache_hit_count =\
                    # 468, block_read_count = 580, block_read_byte = 445, ...
                    token_list = line.strip().split(",")
                    # token_list = ['user_key_comparison_count = 500',
                    # 'block_cache_hit_count = 468','block_read_count = 580'...
                    perf_context = {
                        tk.split("=")[0].strip(): tk.split("=")[1].strip()
                        for tk in token_list
                        if tk
                    }
                    # TODO(poojam23): this is a hack and should be replaced
                    # with the timestamp that db_bench will provide per printed
                    # perf_context
                    timestamp = int(time.time())
                    perf_context_ts = {}
                    for stat in perf_context.keys():
                        perf_context_ts[stat] = {timestamp: int(perf_context[stat])}
                    output[self.PERF_CON] = perf_context_ts
                    perf_context_begins = False
                elif line.startswith(self.DB_PATH):
                    # line from sample output:
                    # DB path: [/tmp/rocksdbtest-155919/dbbench]\n
                    output[self.DB_PATH] = line.split("[")[1].split("]")[0]
        return output

    def get_log_options(self, db_options, db_path):
        # get the location of the LOG file and the frequency at which stats are
        # dumped in the LOG file
        log_dir_path = None
        stats_freq_sec = None
        logs_file_prefix = None

        # fetch frequency at which the stats are dumped in the Rocksdb logs
        dump_period = "DBOptions.stats_dump_period_sec"
        # fetch the directory, if specified, in which the Rocksdb logs are
        # dumped, by default logs are dumped in same location as database
        log_dir = "DBOptions.db_log_dir"
        log_options = db_options.get_options([dump_period, log_dir])
        if dump_period in log_options:
            stats_freq_sec = int(log_options[dump_period][NO_COL_FAMILY])
        if log_dir in log_options:
            log_dir_path = log_options[log_dir][NO_COL_FAMILY]

        log_file_name = DBBenchRunner.get_info_log_file_name(log_dir_path, db_path)

        if not log_dir_path:
            log_dir_path = db_path
        if not log_dir_path.endswith("/"):
            log_dir_path += "/"

        logs_file_prefix = log_dir_path + log_file_name
        return (logs_file_prefix, stats_freq_sec)

    def _get_options_command_line_args_str(self, curr_options):
        """
        This method uses the provided Rocksdb OPTIONS to create a string of
        command-line arguments for db_bench.
        The --options_file argument is always given and the options that are
        not supported by the OPTIONS file are given as separate arguments.
        """
        optional_args_str = DBBenchRunner.get_opt_args_str(
            curr_options.get_misc_options()
        )
        # generate an options configuration file
        options_file = curr_options.generate_options_config(nonce="12345")
        optional_args_str += " --options_file=" + options_file
        return optional_args_str

    def _setup_db_before_experiment(self, curr_options, db_path):
        # remove destination directory if it already exists
        try:
            shutil.rmtree(db_path, ignore_errors=True)
        except OSError as e:
            print("Error: rmdir " + e.filename + " " + e.strerror)
        # setup database with a million keys using the fillrandom benchmark
        command = "{} --benchmarks=fillrandom --db={} --num=1000000".format(
            self.db_bench_binary,
            db_path,
        )
        args_str = self._get_options_command_line_args_str(curr_options)
        command += args_str
        self._run_command(command)

    def _build_experiment_command(self, curr_options, db_path):
        command = "{} --benchmarks={} --statistics --perf_level=3 --db={}".format(
            self.db_bench_binary,
            self.benchmark,
            db_path,
        )
        # fetch the command-line arguments string for providing Rocksdb options
        args_str = self._get_options_command_line_args_str(curr_options)
        # handle the command-line args passed in the constructor, these
        # arguments are specific to db_bench
        for cmd_line_arg in self.db_bench_args:
            args_str += " --" + cmd_line_arg
        command += args_str
        return command

    def _run_command(self, command):
        out_file = open(self.OUTPUT_FILE, "w+")
        err_file = open(self.ERROR_FILE, "w+")
        print("executing... - " + command)
        subprocess.call(command, shell=True, stdout=out_file, stderr=err_file)
        out_file.close()
        err_file.close()

    def run_experiment(self, db_options, db_path):
        # setup the Rocksdb database before running experiment
        self._setup_db_before_experiment(db_options, db_path)
        # get the command to run the experiment
        command = self._build_experiment_command(db_options, db_path)
        experiment_start_time = int(time.time())
        # run experiment
        self._run_command(command)
        experiment_end_time = int(time.time())
        # parse the db_bench experiment output
        parsed_output = self._parse_output(get_perf_context=True)

        # get the log files path prefix and frequency at which Rocksdb stats
        # are dumped in the logs
        logs_file_prefix, stats_freq_sec = self.get_log_options(
            db_options, parsed_output[self.DB_PATH]
        )
        # create the Rocksbd LOGS object
        db_logs = DatabaseLogs(logs_file_prefix, db_options.get_column_families())
        # Create the Log STATS object
        db_log_stats = LogStatsParser(logs_file_prefix, stats_freq_sec)
        # Create the PerfContext STATS object
        db_perf_context = DatabasePerfContext(parsed_output[self.PERF_CON], 0, False)
        # create the data-sources dictionary
        data_sources = {
            DataSource.Type.DB_OPTIONS: [db_options],
            DataSource.Type.LOG: [db_logs],
            DataSource.Type.TIME_SERIES: [db_log_stats, db_perf_context],
        }
        # Create the ODS STATS object
        if self.ods_args:
            key_prefix = ""
            if "key_prefix" in self.ods_args:
                key_prefix = self.ods_args["key_prefix"]
            data_sources[DataSource.Type.TIME_SERIES].append(
                OdsStatsFetcher(
                    self.ods_args["client_script"],
                    self.ods_args["entity"],
                    experiment_start_time,
                    experiment_end_time,
                    key_prefix,
                )
            )
        # return the experiment's data-sources and throughput
        return data_sources, parsed_output[self.THROUGHPUT]



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import os
import unittest

from advisor.db_log_parser import NO_COL_FAMILY
from advisor.db_options_parser import DatabaseOptions
from advisor.rule_parser import Condition, OptionCondition


class TestDatabaseOptions(unittest.TestCase):
    def setUp(self):
        self.this_path = os.path.abspath(os.path.dirname(__file__))
        self.og_options = os.path.join(self.this_path, "input_files/OPTIONS-000005")
        misc_options = ["bloom_bits = 4", "rate_limiter_bytes_per_sec = 1024000"]
        # create the options object
        self.db_options = DatabaseOptions(self.og_options, misc_options)
        # perform clean-up before running tests
        self.generated_options = os.path.join(
            self.this_path, "../temp/OPTIONS_testing.tmp"
        )
        if os.path.isfile(self.generated_options):
            os.remove(self.generated_options)

    def test_get_options_diff(self):
        old_opt = {
            "DBOptions.stats_dump_freq_sec": {NO_COL_FAMILY: "20"},
            "CFOptions.write_buffer_size": {
                "default": "1024000",
                "col_fam_A": "128000",
                "col_fam_B": "128000000",
            },
            "DBOptions.use_fsync": {NO_COL_FAMILY: "true"},
            "DBOptions.max_log_file_size": {NO_COL_FAMILY: "128000000"},
        }
        new_opt = {
            "bloom_bits": {NO_COL_FAMILY: "4"},
            "CFOptions.write_buffer_size": {
                "default": "128000000",
                "col_fam_A": "128000",
                "col_fam_C": "128000000",
            },
            "DBOptions.use_fsync": {NO_COL_FAMILY: "true"},
            "DBOptions.max_log_file_size": {NO_COL_FAMILY: "0"},
        }
        diff = DatabaseOptions.get_options_diff(old_opt, new_opt)

        expected_diff = {
            "DBOptions.stats_dump_freq_sec": {NO_COL_FAMILY: ("20", None)},
            "bloom_bits": {NO_COL_FAMILY: (None, "4")},
            "CFOptions.write_buffer_size": {
                "default": ("1024000", "128000000"),
                "col_fam_B": ("128000000", None),
                "col_fam_C": (None, "128000000"),
            },
            "DBOptions.max_log_file_size": {NO_COL_FAMILY: ("128000000", "0")},
        }
        self.assertDictEqual(diff, expected_diff)

    def test_is_misc_option(self):
        self.assertTrue(DatabaseOptions.is_misc_option("bloom_bits"))
        self.assertFalse(
            DatabaseOptions.is_misc_option("DBOptions.stats_dump_freq_sec")
        )

    def test_set_up(self):
        options = self.db_options.get_all_options()
        self.assertEqual(22, len(options.keys()))
        expected_misc_options = {
            "bloom_bits": "4",
            "rate_limiter_bytes_per_sec": "1024000",
        }
        self.assertDictEqual(expected_misc_options, self.db_options.get_misc_options())
        self.assertListEqual(
            ["default", "col_fam_A"], self.db_options.get_column_families()
        )

    def test_get_options(self):
        opt_to_get = [
            "DBOptions.manual_wal_flush",
            "DBOptions.db_write_buffer_size",
            "bloom_bits",
            "CFOptions.compaction_filter_factory",
            "CFOptions.num_levels",
            "rate_limiter_bytes_per_sec",
            "TableOptions.BlockBasedTable.block_align",
            "random_option",
        ]
        options = self.db_options.get_options(opt_to_get)
        expected_options = {
            "DBOptions.manual_wal_flush": {NO_COL_FAMILY: "false"},
            "DBOptions.db_write_buffer_size": {NO_COL_FAMILY: "0"},
            "bloom_bits": {NO_COL_FAMILY: "4"},
            "CFOptions.compaction_filter_factory": {
                "default": "nullptr",
                "col_fam_A": "nullptr",
            },
            "CFOptions.num_levels": {"default": "7", "col_fam_A": "5"},
            "rate_limiter_bytes_per_sec": {NO_COL_FAMILY: "1024000"},
            "TableOptions.BlockBasedTable.block_align": {
                "default": "false",
                "col_fam_A": "true",
            },
        }
        self.assertDictEqual(expected_options, options)

    def test_update_options(self):
        # add new, update old, set old
        # before updating
        expected_old_opts = {
            "DBOptions.db_log_dir": {NO_COL_FAMILY: None},
            "DBOptions.manual_wal_flush": {NO_COL_FAMILY: "false"},
            "bloom_bits": {NO_COL_FAMILY: "4"},
            "CFOptions.num_levels": {"default": "7", "col_fam_A": "5"},
            "TableOptions.BlockBasedTable.block_restart_interval": {"col_fam_A": "16"},
        }
        get_opts = list(expected_old_opts.keys())
        options = self.db_options.get_options(get_opts)
        self.assertEqual(expected_old_opts, options)
        # after updating options
        update_opts = {
            "DBOptions.db_log_dir": {NO_COL_FAMILY: "/dev/shm"},
            "DBOptions.manual_wal_flush": {NO_COL_FAMILY: "true"},
            "bloom_bits": {NO_COL_FAMILY: "2"},
            "CFOptions.num_levels": {"col_fam_A": "7"},
            "TableOptions.BlockBasedTable.block_restart_interval": {"default": "32"},
            "random_misc_option": {NO_COL_FAMILY: "something"},
        }
        self.db_options.update_options(update_opts)
        update_opts["CFOptions.num_levels"]["default"] = "7"
        update_opts["TableOptions.BlockBasedTable.block_restart_interval"] = {
            "default": "32",
            "col_fam_A": "16",
        }
        get_opts.append("random_misc_option")
        options = self.db_options.get_options(get_opts)
        self.assertDictEqual(update_opts, options)
        expected_misc_options = {
            "bloom_bits": "2",
            "rate_limiter_bytes_per_sec": "1024000",
            "random_misc_option": "something",
        }
        self.assertDictEqual(expected_misc_options, self.db_options.get_misc_options())

    def test_generate_options_config(self):
        # make sure file does not exist from before
        self.assertFalse(os.path.isfile(self.generated_options))
        self.db_options.generate_options_config("testing")
        self.assertTrue(os.path.isfile(self.generated_options))

    def test_check_and_trigger_conditions(self):
        # options only from CFOptions
        # setup the OptionCondition objects to check and trigger
        update_dict = {
            "CFOptions.level0_file_num_compaction_trigger": {"col_fam_A": "4"},
            "CFOptions.max_bytes_for_level_base": {"col_fam_A": "10"},
        }
        self.db_options.update_options(update_dict)
        cond1 = Condition("opt-cond-1")
        cond1 = OptionCondition.create(cond1)
        cond1.set_parameter(
            "options",
            [
                "CFOptions.level0_file_num_compaction_trigger",
                "TableOptions.BlockBasedTable.block_restart_interval",
                "CFOptions.max_bytes_for_level_base",
            ],
        )
        cond1.set_parameter(
            "evaluate", "int(options[0])*int(options[1])-int(options[2])>=0"
        )
        # only DBOptions
        cond2 = Condition("opt-cond-2")
        cond2 = OptionCondition.create(cond2)
        cond2.set_parameter(
            "options",
            [
                "DBOptions.db_write_buffer_size",
                "bloom_bits",
                "rate_limiter_bytes_per_sec",
            ],
        )
        cond2.set_parameter(
            "evaluate", "(int(options[2]) * int(options[1]) * int(options[0]))==0"
        )
        # mix of CFOptions and DBOptions
        cond3 = Condition("opt-cond-3")
        cond3 = OptionCondition.create(cond3)
        cond3.set_parameter(
            "options",
            [
                "DBOptions.db_write_buffer_size",  # 0
                "CFOptions.num_levels",  # 5, 7
                "bloom_bits",  # 4
            ],
        )
        cond3.set_parameter(
            "evaluate", "int(options[2])*int(options[0])+int(options[1])>6"
        )
        self.db_options.check_and_trigger_conditions([cond1, cond2, cond3])

        cond1_trigger = {"col_fam_A": ["4", "16", "10"]}
        self.assertDictEqual(cond1_trigger, cond1.get_trigger())
        cond2_trigger = {NO_COL_FAMILY: ["0", "4", "1024000"]}
        self.assertDictEqual(cond2_trigger, cond2.get_trigger())
        cond3_trigger = {"default": ["0", "7", "4"]}
        self.assertDictEqual(cond3_trigger, cond3.get_trigger())


if __name__ == "__main__":
    unittest.main()
